# jeff65 gold-syntax lowering passes
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from . import asm
from ... import ast, pattern
from ...pattern import Predicate as P


@pattern.transform(pattern.Order.Any)
class LowerAssignment:
    @pattern.match(
        ast.AstNode("block", {
            "stmt": ast.AstNode("set", span=P("span"), attrs={
                "type": P("ty"),
                "lvalue": P("lvalue"),
                "rvalue": P("rvalue"),
            }),
            "next": P("nxt"),
        }))
    def lower_set(self, span, ty, lvalue, rvalue, nxt):
        assert ty.width == lvalue.attrs["width"]
        assert ty.width == rvalue.attrs["width"]
        return ast.AstNode.make_sequence("block", "stmt", [
            asm.lda(rvalue, span),
            asm.sta(lvalue, span),
        ], rest=nxt)


class LowerFunctions(ast.TranslationPass):
    def exit_fun(self, node):
        children = node.select("body", "stmt")
        children.append(asm.rts(node.span))
        return node.update_attrs({
            "body": ast.AstNode.make_sequence("block", "stmt", children),
        })



# jeff65 gold-syntax type-related passes
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from . import binding
from ... import ast
from ...blum import types


class ConstructTypes(ast.TranslationPass):
    builtin_types = {
        'u8': types.u8, 'u16': types.u16, 'u24': types.u24, 'u32': types.u32,
        'i8': types.i8, 'i16': types.i16, 'i24': types.i24, 'i32': types.i32,
    }

    def enter_unit(self, node):
        # TODO scoping for when we import types from other units
        self.known_types = self.builtin_types
        return node

    def enter_type_ref(self, node):
        inner_type = self.known_types[node.attrs['type']]
        return types.RefType(inner_type)


class PropagateTypes(binding.ScopedPass):
    def enter_identifier(self, node):
        t = self.look_up_name(node.attrs['name'])
        return node.update_attrs({"type": t})

    def exit_deref(self, node):
        return node.update_attrs({
            "type": node.select("address", "type")[0].target,
        })

    def exit_set(self, node):
        return node.update_attrs({
            "type": node.select("lvalue", "type")[0],
        })

    def enter_fun(self, node):
        attrs = node.attrs.asbuilder()
        attrs['type'] = types.FunctionType(
            node.attrs['return'] or types.void,
            *node.select("args", "arg"))
        del attrs['return']
        return node.replace_attrs(attrs)



# jeff65 gold-syntax 'mem' virtual unit
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from ..blum import types
from . import units


class MemUnit(units.ExternalUnit):
    """The 'mem' virtual unit.

    This is a built-in virtual unit providing intrinsics for performing direct
    memory access.
    """

    def __init__(self):
        super().__init__("mem")
        self.members = {}
        self.add_member('as-pointer', self._as_pointer)
        self.add_member('as-address', self._as_address)

    def add_member(self, name, impl):
        self.members[name] = units.IntrinsicSymbol(
            name, self, impl.intrinsic_type, impl)

    def member(self, name):
        return self.members[name]

    @units.member_type(types.FunctionType(types.ptr, types.u16))
    def _as_pointer(self, arg):
        # All this does is pass the result back out again -- it's just to get
        # around the type checker.
        return arg

    @units.member_type(types.FunctionType(types.u16, types.ptr))
    def _as_address(self, arg):
        # All this does is pass the result back out again -- it's just to get
        # around the type checker.
        return arg

    def __repr__(self):
        return "MemUnit()"



# jeff65 gold-syntax grammar
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import enum
import re
from ..parsing import Grammar, Lexer, Parser, ReStream, Rule


@enum.unique
class Mode(enum.IntEnum):
    NORMAL = Parser.NORMAL_MODE
    COMMENT = 1
    STRING = 2


T = enum.Enum('T', [
    # control tokens -- these tend to cause lexer mode switches
    'EOF', 'MYSTERY', 'STRING_DELIM', 'COMMENT_OPEN', 'COMMENT_CLOSE',

    # literals tokens
    'IDENTIFIER', 'NUMERIC', 'STRING', 'STRING_ESCAPE', 'WHITESPACE',
    'COMMENT_TEXT',

    # arithmetic operators
    'OPERATOR_PLUS', 'OPERATOR_MINUS', 'OPERATOR_TIMES', 'OPERATOR_DIVIDE',

    # logical operators
    'OPERATOR_NOT', 'OPERATOR_AND', 'OPERATOR_OR',

    # bitwise operators
    'OPERATOR_SHR', 'OPERATOR_SHL', 'OPERATOR_BITNOT', 'OPERATOR_BITAND',
    'OPERATOR_BITOR', 'OPERATOR_BITXOR',

    # comparison operators
    'OPERATOR_NE', 'OPERATOR_EQ', 'OPERATOR_LE', 'OPERATOR_GE', 'OPERATOR_LT',
    'OPERATOR_GT',

    # assignment operators
    'OPERATOR_ASSIGN', 'OPERATOR_ASSIGN_INC', 'OPERATOR_ASSIGN_DEC',

    # member access operators
    'OPERATOR_DOT',

    # pointer operators
    'OPERATOR_DEREF', 'OPERATOR_REF',

    # statement keywords
    'STMT_CONSTANT', 'STMT_FOR', 'STMT_FUN', 'STMT_IF', 'STMT_ISR', 'STMT_LET',
    'STMT_RETURN', 'STMT_USE', 'STMT_WHILE',

    # storage classes
    'STORAGE_MUT', 'STORAGE_STASH',

    # assorted punctuation
    'PUNCT_DO', 'PUNCT_ELSE', 'PUNCT_ELSEIF', 'PUNCT_END', 'PUNCT_ENDFUN',
    'PUNCT_ENDISR', 'PUNCT_IN', 'PUNCT_THEN', 'PUNCT_TO', 'PUNCT_COLON',
    'PUNCT_SEMICOLON', 'PUNCT_COMMA', 'PUNCT_ARROWR',

    # delimiters
    'PAREN_OPEN', 'PAREN_CLOSE', 'BRACKET_OPEN', 'BRACKET_CLOSE', 'BRACE_OPEN',
    'BRACE_CLOSE',
])


# precedences. Note that these are in order of lowest precedence to highest
# precedence. Has to be an IntEnum for comparisons to work.
P = enum.IntEnum('P', [
    # Statements
    'STATEMENTS',
    'RETURN_VALUE',
    'ASSIGNMENTS',

    # Statement elements
    'TYPES',
    'STORAGE',

    # Expressions
    'LITERALS',
    'COMPARISONS',
    'SUMS',
    'PRODUCTS',
    'BITSHIFTS',
    'BITXOR',
    'BITOR',
    'BITAND',
    'PREFIX_UNARY',
    'CALLS',
    'SUBSCRIPTS',
    'MEMBERS',
    'PARENTHESES',
])


# non-whitespace characters which can end tokens
specials = re.escape(r'()[]{}:;.,"\@&')


# creates a regex that matches the given word as long as it is followed by
# something that can end a token. Useful for defining keywords.
def _w(word):
    return r'(?m){}(?=[\s{}]|$)'.format(re.escape(word), specials)


lex = Lexer(T.EOF, [
    # whitespace -> hidden channel
    (Mode.NORMAL, r'(?m)\s+', T.WHITESPACE, ReStream.CHANNEL_HIDDEN),

    # keywords. Must come before the identifier match
    (_w('and'), T.OPERATOR_AND),
    (_w('bitand'), T.OPERATOR_BITAND),
    (_w('bitor'), T.OPERATOR_BITOR),
    (_w('bitxor'), T.OPERATOR_BITXOR),
    (_w('constant'), T.STMT_CONSTANT),
    (_w('do'), T.PUNCT_DO),
    (_w('else'), T.PUNCT_ELSE),
    (_w('elseif'), T.PUNCT_ELSEIF),
    (_w('end'), T.PUNCT_END),
    (_w('endfun'), T.PUNCT_ENDFUN),
    (_w('endisr'), T.PUNCT_ENDISR),
    (_w('for'), T.STMT_FOR),
    (_w('fun'), T.STMT_FUN),
    (_w('if'), T.STMT_IF),
    (_w('in'), T.PUNCT_IN),
    (_w('isr'), T.STMT_ISR),
    (_w('let'), T.STMT_LET),
    (_w('mut'), T.STORAGE_MUT),
    (_w('not'), T.OPERATOR_NOT),
    (_w('or'), T.OPERATOR_OR),
    (_w('return'), T.STMT_RETURN),
    (_w('stash'), T.STORAGE_STASH),
    (_w('then'), T.PUNCT_THEN),
    (_w('to'), T.PUNCT_TO),
    (_w('use'), T.STMT_USE),
    (_w('while'), T.STMT_WHILE),

    # Numeric tokens. Must come before the word match
    (r'\d[^\s{}]*'.format(specials), T.NUMERIC),

    # Identifiers. Matches a letter, followed by zero or more non-token-ending
    # characters. As written, this will actually match numbers as well, but
    # because that one is run first we don't have to worry about that.
    (r'\w[^\s{}]*'.format(specials), T.IDENTIFIER),

    # comment opener. When the lexer comes back, it will be in comment mode
    (Mode.NORMAL, re.escape('/*'), T.COMMENT_OPEN, ReStream.CHANNEL_HIDDEN),

    # comment delimiers, but for comment mode.
    (Mode.COMMENT, re.escape('/*'), T.COMMENT_OPEN, ReStream.CHANNEL_HIDDEN),
    (Mode.COMMENT, re.escape('*/'), T.COMMENT_CLOSE, ReStream.CHANNEL_HIDDEN),

    # This is necessary because the next pattern matches up to, but not
    # including, the newline; however, it will happily match zero characters,
    # causing an infinite loop. This matches that last newline.
    (Mode.COMMENT, r'\n', T.COMMENT_TEXT, ReStream.CHANNEL_HIDDEN),

    # Matches either to the next comment-control token, or the end of the line,
    # whichever happens first.
    (Mode.COMMENT, r'.*?(?=\/\*|\*\/|$)', T.COMMENT_TEXT,
     ReStream.CHANNEL_HIDDEN),

    # String delimiter. When the lexer comes back, it will be in string mode
    (re.escape('"'), T.STRING_DELIM),

    # String control tokens
    (Mode.STRING, r'\\.', T.STRING_ESCAPE),
    (Mode.STRING, re.escape('"'), T.STRING_DELIM),

    # Matches non-special text inside a string. The newline-matching pattern is
    # for the same reason as for comments.
    (Mode.STRING, r'\n', T.STRING),
    (Mode.STRING, r'.*?(?=\\|"|$)', T.STRING),

    # operators & punctuation. These must be ordered such that if A is a prefix
    # of B, then B comes before A. The easiest way to do this is to order them
    # by length.
    (re.escape('->'), T.PUNCT_ARROWR),
    (re.escape('>>'), T.OPERATOR_SHR),
    (re.escape('<<'), T.OPERATOR_SHL),
    (re.escape('!='), T.OPERATOR_NE),
    (re.escape('=='), T.OPERATOR_EQ),
    (re.escape('<='), T.OPERATOR_LE),
    (re.escape('>='), T.OPERATOR_GE),
    (re.escape('+='), T.OPERATOR_ASSIGN_INC),
    (re.escape('-='), T.OPERATOR_ASSIGN_DEC),
    (re.escape('+'), T.OPERATOR_PLUS),
    (re.escape('-'), T.OPERATOR_MINUS),
    (re.escape('*'), T.OPERATOR_TIMES),
    (re.escape('/'), T.OPERATOR_DIVIDE),
    (re.escape('<'), T.OPERATOR_LT),
    (re.escape('>'), T.OPERATOR_GT),
    (re.escape('='), T.OPERATOR_ASSIGN),
    (re.escape('.'), T.OPERATOR_DOT),
    (re.escape('@'), T.OPERATOR_DEREF),
    (re.escape('&'), T.OPERATOR_REF),
    (re.escape(':'), T.PUNCT_COLON),
    (re.escape(';'), T.PUNCT_SEMICOLON),
    (re.escape(','), T.PUNCT_COMMA),
    (re.escape('('), T.PAREN_OPEN),
    (re.escape(')'), T.PAREN_CLOSE),
    (re.escape('['), T.BRACKET_OPEN),
    (re.escape(']'), T.BRACKET_CLOSE),
    (re.escape('{'), T.BRACE_OPEN),
    (re.escape('}'), T.BRACE_CLOSE),

    # If we fail to match anything, consume one character, and move on.
    (r'.', T.MYSTERY),
])


grammar = Grammar('start', [T.EOF], [
    Rule('alist_inner', ['expr']),
    Rule('alist_inner', ['expr', T.PUNCT_COMMA, 'alist_inner']),
    Rule('alist', []),
    Rule('alist', ['alist_inner']),

    Rule('member', [T.IDENTIFIER]),

    Rule('expr', [T.PAREN_OPEN, 'expr', T.PAREN_CLOSE], prec=P.PARENTHESES),
    Rule('expr', ['expr', T.OPERATOR_DOT, 'member'], prec=P.MEMBERS),
    Rule('expr', ['expr', T.BRACKET_OPEN, 'expr', T.BRACKET_CLOSE],
         prec=P.SUBSCRIPTS),
    Rule('expr', ['stmt_call'], prec=P.CALLS),
    Rule('expr', [(T.OPERATOR_DEREF,
                   T.OPERATOR_MINUS,
                   T.OPERATOR_BITNOT), 'expr'],
         prec=P.PREFIX_UNARY,
         rassoc=True),
    Rule('expr', ['expr', T.OPERATOR_BITAND, 'expr'], prec=P.BITAND),
    Rule('expr', ['expr', T.OPERATOR_BITOR, 'expr'], prec=P.BITOR),
    Rule('expr', ['expr', T.OPERATOR_BITXOR, 'expr'], prec=P.BITXOR),
    Rule('expr', ['expr', (T.OPERATOR_SHL,
                           T.OPERATOR_SHR), 'expr'], prec=P.BITSHIFTS),
    Rule('expr', ['expr', (T.OPERATOR_TIMES,
                           T.OPERATOR_DIVIDE), 'expr'], prec=P.PRODUCTS),
    Rule('expr', ['expr', (T.OPERATOR_PLUS,
                           T.OPERATOR_MINUS), 'expr'], prec=P.SUMS),
    Rule('expr', ['expr', (T.OPERATOR_EQ, T.OPERATOR_NE,
                           T.OPERATOR_LE, T.OPERATOR_GE,
                           T.OPERATOR_LT, T.OPERATOR_GT), 'expr'],
         prec=P.COMPARISONS),
    Rule('expr', [(T.NUMERIC, T.IDENTIFIER, 'string')], prec=P.LITERALS),

    Rule('array', [T.BRACKET_OPEN, 'alist', T.BRACKET_CLOSE]),

    Rule('string', [T.STRING_DELIM, 'string_inner', T.STRING_DELIM],
         prec=P.LITERALS),
    Rule('string_inner', [], mode=Mode.STRING),
    Rule('string_inner', ['string_inner', (T.STRING, T.STRING_ESCAPE)],
         mode=Mode.STRING),

    Rule('storage', [], prec=P.STORAGE),
    Rule('storage', [(T.STORAGE_MUT, T.STORAGE_STASH)]),

    Rule('range_to', ['expr', T.PUNCT_TO, 'expr']),

    Rule('type_id', [T.IDENTIFIER]),
    Rule('type_id', [
        T.OPERATOR_REF, T.BRACKET_OPEN,
        'storage', 'type_id', T.BRACKET_CLOSE], prec=P.TYPES),
    Rule('type_id', [T.OPERATOR_REF, 'storage', 'type_id']),
    Rule('type_id', [T.BRACKET_OPEN, 'storage', 'type_id', T.PUNCT_SEMICOLON,
                     ('expr', 'range_to'), T.BRACKET_CLOSE]),

    Rule('declaration', [T.IDENTIFIER, T.PUNCT_COLON, 'type_id']),

    Rule('stmt_constant', [T.STMT_CONSTANT, 'declaration',
                           T.OPERATOR_ASSIGN, ('expr', 'array')],
         prec=P.STATEMENTS),

    Rule('stmt_use', [T.STMT_USE, T.IDENTIFIER]),

    Rule('stmt_let', [T.STMT_LET, 'storage', 'declaration',
                      T.OPERATOR_ASSIGN, ('expr', 'array')],
         prec=P.STATEMENTS),

    Rule('do_block', [T.PUNCT_DO, 'block', T.PUNCT_END]),

    Rule('stmt_while', [T.STMT_WHILE, 'expr', 'do_block']),

    Rule('stmt_for', [T.STMT_FOR, 'declaration',
                      T.PUNCT_IN, ('range_to', 'expr'), 'do_block']),

    Rule('branch_else_if', [T.PUNCT_ELSEIF, 'expr', T.PUNCT_THEN, 'block']),
    Rule('branch_else', [T.PUNCT_ELSE, 'block']),
    Rule('branch_else_ifs', ['branch_else_ifs', 'branch_else_if']),
    Rule('branch_else_ifs', ['branch_else_if']),
    Rule('stmt_if', [T.STMT_IF, 'expr', T.PUNCT_THEN, 'block', T.PUNCT_END]),
    Rule('stmt_if', [T.STMT_IF, 'expr', T.PUNCT_THEN, 'block',
                     'branch_else', T.PUNCT_END]),
    Rule('stmt_if', [T.STMT_IF, 'expr', T.PUNCT_THEN, 'block',
                     'branch_else_ifs', 'branch_else', T.PUNCT_END]),

    Rule('stmt_isr', [T.STMT_ISR, T.IDENTIFIER, 'block', T.PUNCT_ENDISR]),

    Rule('plist', []),
    Rule('plist', ['plist_inner']),
    Rule('plist_inner', ['declaration']),
    Rule('plist_inner', ['plist_inner', T.PUNCT_COMMA, 'declaration']),
    Rule('stmt_fun', [T.STMT_FUN, T.IDENTIFIER,
                      T.PAREN_OPEN, 'plist', T.PAREN_CLOSE,
                      'block', T.PUNCT_ENDFUN]),
    Rule('stmt_fun', [T.STMT_FUN, T.IDENTIFIER,
                      T.PAREN_OPEN, 'plist', T.PAREN_CLOSE,
                      T.PUNCT_ARROWR, 'type_id',
                      'block', T.PUNCT_ENDFUN]),

    Rule('stmt_return', [T.STMT_RETURN], prec=P.STATEMENTS),
    Rule('stmt_return', [T.STMT_RETURN, 'expr'], prec=P.RETURN_VALUE),

    Rule('stmt_assign', ['expr', (T.OPERATOR_ASSIGN,
                                  T.OPERATOR_ASSIGN_INC,
                                  T.OPERATOR_ASSIGN_DEC), 'expr'],
         prec=P.ASSIGNMENTS),

    Rule('stmt_call', ['expr', T.PAREN_OPEN, 'alist', T.PAREN_CLOSE],
         prec=P.CALLS),

    Rule('block', [], prec=P.STATEMENTS, rassoc=True),
    Rule('block', [('stmt_constant',
                    'stmt_for',
                    'stmt_if',
                    'stmt_let',
                    'stmt_return',
                    'stmt_while',
                    'stmt_assign',
                    'stmt_call'),
                   'block'],
         prec=P.STATEMENTS,
         rassoc=True),

    Rule('toplevel', []),
    Rule('toplevel', [('stmt_constant',
                       'stmt_isr',
                       'stmt_let',
                       'stmt_use',
                       'stmt_fun'),
                      'toplevel']),

    Rule('unit', ['toplevel']),
    Rule('start', ['unit']),
])

comment_grammar = Grammar('start', [T.COMMENT_CLOSE, T.WHITESPACE, T.EOF], [
    # left-recursive definition of a comment
    Rule('comment_inner', [], mode=Mode.COMMENT),
    Rule('comment_inner', ['comment_inner',
                           (T.COMMENT_TEXT, 'comment_nested')],
         mode=Mode.COMMENT),

    # we have to define a rule for nested comments separately to avoid
    # mode/mode conflicts
    Rule('comment_nested', [T.COMMENT_OPEN, 'comment_inner', T.COMMENT_CLOSE],
         mode=Mode.COMMENT),

    # enter/exit mode split
    Rule('comment', [T.COMMENT_OPEN, 'comment_inner', T.COMMENT_CLOSE]),

    # the empty rule handles whitespace
    Rule('hidden', []),
    Rule('hidden', ['comment']),
    Rule('start', ['hidden']),
])

parse = grammar.build_parser({
    ReStream.CHANNEL_HIDDEN: comment_grammar,
})



# jeff65 gold-syntax unit resolution
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


class ExternalUnit:
    """Represents an external unit."""

    def __init__(self, source):
        self.source = source

    def member(self, name):
        """Gets a member of a unit by name."""
        pass

    def __repr__(self):
        return "ExternalUnit({})".format(repr(self.source))


class UnitSymbol:
    """Represents a symbol exported from a unit."""

    def __init__(self, name, unit, t, is_intrinsic=False):
        self.name = name
        self.unit = unit
        self.type = t
        self.is_intrinsic = is_intrinsic

    def __repr__(self):
        return "<{} {}.{}: {}>".format(
            type(self).__name__, self.unit, self.name, self.type)


class IntrinsicSymbol(UnitSymbol):
    """Represents a compiler intrinsic symbol.

    Compiler intrinsic symbols receive their arguments as AST subtrees at
    translation time, and output a new AST subtree based on them. Intrinsics
    are provided by virtual units.
    """

    def __init__(self, name, unit, t, impl):
        super().__init__(name, unit, t, is_intrinsic=True)
        self.impl = impl

    def __call__(self, *args):
        return self.impl(*args)


def member_type(t):
    def add_to_fun(fun):
        fun.intrinsic_type = t
        return fun

    return add_to_fun



# jeff65 pattern-based AST transformation
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Pattern-based AST transformation system.

This module provides something similar to Scheme R5RS syntax-rules, on top of
procedural DFS transform system.
"""

import attr
import enum
import inspect
from . import ast


class Order(enum.Enum):
    """Denotes the order that nodes are matched.

    'Descending' results in nodes being matched from outermost to innermost,
    i.e. during the descent of the DFS traversal. 'Ascending' results in the
    opposite.

    Use 'Any' if the order does not matter for the transform to work correctly.
    """
    Descending = 0
    Ascending = 1

    # given a choice, prefer descending, so that if the transform prunes the
    # tree, the pruned nodes won't have to be traversed.
    Any = Descending


_marker = object()


def match(pattern):
    """Marks a method as a pattern transformer."""

    return lambda method: (_marker, pattern, method)


def transform(order):
    """Converts the decorated class into a transform."""

    def _decorate_transform(cls):
        analyser = PatternAnalyser()
        ptpairs = []
        m_dict = {'ptpairs': ptpairs}

        if order == Order.Descending:
            m_dict['transform_enter'] = transform_handler
            m_dict['transform_exit'] = dummy_handler
        else:
            m_dict['transform_enter'] = dummy_handler
            m_dict['transform_exit'] = transform_handler

        for member, value in vars(cls).items():
            if not (isinstance(value, tuple)
                    and len(value) == 3
                    and value[0] == _marker):
                # copy across non-pattern stuff unchanged
                m_dict[member] = value
            else:
                _, pattern, template = value
                if isinstance(pattern, ast.AstNode):
                    predicates = pattern.transform(analyser)
                else:
                    # do the non-recursive transform
                    predicates = analyser.make_predicate(pattern)
                m_dict[member] = template
                ptpairs.append((predicates, template))
        return type(cls.__name__, cls.__bases__, m_dict)

    return _decorate_transform


class MatchError(Exception):
    pass


def dummy_handler(self, t, node):
    return node


def transform_handler(self, t, node):
    for predicate, template in self.ptpairs:
        captures = {}
        if predicate._match(node, captures):
            f = template.__get__(self, type)
            n = f(**captures)
            if isinstance(n, ast.AstNode) and n.span is None:
                return attr.evolve(n, span=node.span)
            return n
    return node


class PatternAnalyser:
    """Converts a pattern into a bound predicate."""

    def make_predicate(self, obj):
        if isinstance(obj, Predicate):
            # objects which are already predicates are just passed on
            return obj
        # fallback case: check for equality without capturing.
        return Predicate.eq(obj)

    def make_span_predicate(self, span):
        if span is None:
            return Predicate.any()
        return self.make_predicate(span)

    def make_attrs_predicate(self, attrs):
        if isinstance(attrs, Predicate):
            return attrs

        pas = {}
        exhaustive = False
        for k, v in attrs.items():
            if k == "exhaustive!":
                # yeah, the in-band signaling is a little weird, but I'm not
                # really sure where else to put this flag.
                exhaustive = v
            else:
                pas[k] = self.make_predicate(v)
        keys = set(pas.keys())

        def _attrs_predicate(attrs, captures):
            if exhaustive and set(attrs.keys()) != keys:
                return False
            for k, v in pas.items():
                if not v._match(attrs[k], captures):
                    return False
            return True
        return Predicate(None, _attrs_predicate)

    def transform_enter(self, t, node):
        return node

    def transform_exit(self, t, node):
        return Predicate.node(
            self.make_predicate(node.t),
            self.make_attrs_predicate(node.attrs),
            self.make_span_predicate(node.span))


class Predicate:
    def __init__(self, key, predicate=True):
        self.key = key
        if not callable(predicate):
            self.predicate = lambda _1, _2: predicate
        else:
            self.predicate = predicate

    def _match(self, value, captures):
        if not self.predicate(value, captures):
            return False
        if self.key is not None and not self.key.startswith('!'):
            captures[self.key] = value
        return True

    @classmethod
    def any(cls):
        return cls(None)

    @classmethod
    def any_node(cls, key=None, with_attrs=None):
        if with_attrs is None:
            attrs = cls.any()
        else:
            # this is goofy but the children have to be transformed now or it's
            # not happening...
            analyser = PatternAnalyser()
            attrs = analyser.make_attrs_predicate(with_attrs)

        return cls.node(
            cls.any(),
            attrs,
            cls.any(),
            key=key)

    @classmethod
    def node(cls, pt, pa, pn, key=None):
        def _node_predicate(node, captures):
            return (isinstance(node, ast.AstNode)
                    and pt._match(node.t, captures)
                    and pa._match(node.attrs, captures)
                    and pn._match(node.span, captures))
        return cls(key, _node_predicate)

    @classmethod
    def require(cls, value_or_predicate, exc=None):
        exc = exc or MatchError

        if callable(value_or_predicate):
            predicate = value_or_predicate
            try:
                value = inspect.getsource(value_or_predicate)
            except OSError:
                value = '<predicate>'
        else:
            def predicate(v, c):
                return v == value_or_predicate
            value = value_or_predicate

        def _p_require(v, captures):
            if not predicate(v, captures):
                raise exc(f"Expected {value} got {v}")
            return True
        return cls(None, _p_require)

    @classmethod
    def eq(cls, value, key=None, require=False):
        def _p_eq(v, captures):
            if v == value:
                return True
            if require:
                raise MatchError(f"Expected {value}, got {v}")
            return False
        return cls(key, _p_eq)

    @classmethod
    def lt(cls, value, key=None, require=False):
        def _p_lt(v, captures):
            if v < value:
                return True
            if require:
                raise MatchError(f"Expected value <{value}, got {v}")
            return False
        return cls(key, _p_lt)



# jeff65 parser generator
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import attr
import logging
import re
import time
from itertools import chain

logger = logging.getLogger(__name__)


class ParseError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


@attr.s(slots=True, frozen=True)
class TextSpan:
    start_line = attr.ib()
    start_column = attr.ib()
    end_line = attr.ib()
    end_column = attr.ib()

    def __attrs_post_init__(self):
        assert self.end >= self.start

    @property
    def start(self):
        return (self.start_line, self.start_column)

    @property
    def end(self):
        return (self.end_line, self.end_column)

    def __bool__(self):
        return self.start < self.end

    def __contains__(self, other):
        return (
            isinstance(other, TextSpan)
            and self.start <= other.start
            and other.end <= self.end)

    @staticmethod
    def cover(spans):
        """Return a TextSpan covering all of the given spans.

        The result is the shortest span t such that every given is contained in
        it. Note that because TextSpans are always contiguous, there may exist
        spans which are not contained in any of the given spans, but are
        contained in the cover span.
        """
        return TextSpan(
            *min(s.start for s in spans),
            *max(s.end for s in spans))

    def __str__(self):
        start = f'{self.start_line}:{self.start_column}'
        end = f'{self.end_line}:{self.end_column}'
        return f'{start}-{end}'


@attr.s(slots=True, frozen=True)
class Token:
    t = attr.ib()
    text = attr.ib()
    channel = attr.ib(default=0, cmp=False)
    span = attr.ib(default=None, cmp=False)

    def _pretty(self, indent, no_position):
        i = " " * indent
        return f'{i}{self.t}={self.text!r} {self.span}\n'


class ReStream:
    """Regex-matchable stream."""

    CHANNEL_ALL = -1
    CHANNEL_DEFAULT = 0
    CHANNEL_HIDDEN = 1

    def __init__(self, stream):
        self.stream = iter(stream)
        self.current = None
        self.line = 0
        self.column = 0

        try:
            self.advance_line()
        except StopIteration:
            self.current = ""
            self.line = 1

    def advance_line(self):
        """Advance the stream position to the beginning of the next line."""
        try:
            self.current = next(self.stream)
        except StopIteration:
            raise
        else:
            self.line += 1
            self.column = 0

    def assure_line(self):
        """Assures that at least one character remains in the current line."""
        if self.column == len(self.current):
            self.advance_line()

    def match(self, regex):
        """Match the given regex at the current stream position.

        In order to actually advance the position, call ReStream.produce() with
        the returned match object.

        Returns a match object if successful, None otherwise.
        """

        self.assure_line()
        return regex.match(self.current, self.column)

    def produce(self, symbol, match, channel=CHANNEL_DEFAULT):
        """Produce a token and advance the position."""

        token = Token(symbol, match.group(), channel,
                      TextSpan(
                          self.line, self.column,
                          self.line, match.end()))
        self.column = match.end()
        return token

    def rewind(self, token: Token):
        """Rewinds by one token.

        This may only be called if the last method to be called on the ReStream
        object was the produce() call which returned the given token.
        """

        assert token.span.end == (self.line, self.column)
        self.column = token.span.start_column

    def produce_eof(self, symbol):
        """Produce an EOF token."""
        return Token(symbol, None, self.CHANNEL_ALL,
                     TextSpan(self.line, self.column, self.line, self.column))


class Lexer:
    def __init__(self, eof, rules):
        """Create a lexer callable.

        rules should be a list of tuples of one of the following forms:
          (pattern, token_type)
          (mode, pattern, token_type)
          (mode, pattern, token_type, channel)
        """

        self.eof = eof
        self.mode_rules = {}
        for mptc in rules:
            mode, channel = Parser.NORMAL_MODE, ReStream.CHANNEL_DEFAULT
            if len(mptc) == 2:
                pattern, tt = mptc
            elif len(mptc) == 3:
                mode, pattern, tt = mptc
            else:
                mode, pattern, tt, channel = mptc
            rs = self.mode_rules.setdefault(mode, [])
            rs.append((re.compile(pattern), tt, channel))

    def __call__(self, stream: ReStream, mode: int) -> Token:
        try:
            stream.assure_line()
        except StopIteration:
            return stream.produce_eof(self.eof)

        for regex, tt, channel in self.mode_rules[mode]:
            m = stream.match(regex)
            if m:
                return stream.produce(tt, m, channel)
        assert False, "no match!"  # TODO: proper exception


def _convert_lhs(lhs):
    if not isinstance(lhs, Symbol):
        return Symbol(lhs)
    return lhs


def _convert_rhs(rhs):
    result = []
    for sym in rhs:
        if isinstance(sym, tuple):
            # alternations are tuples
            sym = frozenset(_convert_lhs(s) for s in sym)
        elif isinstance(sym, frozenset):
            # already converted
            pass
        else:
            sym = frozenset({_convert_lhs(sym)})
        result.append(sym)
    return tuple(result)


@attr.s(slots=True, frozen=True, repr=False)
class Symbol:
    value = attr.ib()

    @property
    def is_terminal(self):
        # we represent bare nonterminals as strings
        return not isinstance(self.value, str)

    def __repr__(self):
        return f"`{self.value}"

    def extend(self, start, end):
        return ExtendedSymbol(self.value, start, end)

    @property
    def parent(self):
        return Symbol(self.value)


@attr.s(slots=True, frozen=True, repr=False)
class ExtendedSymbol(Symbol):
    start = attr.ib()
    end = attr.ib()

    def __repr__(self):
        return f"`{self.start}_{self.value!r}_{self.end}"


@attr.s(slots=True, frozen=True, repr=False)
class Rule:
    lhs = attr.ib(converter=_convert_lhs)
    rhs = attr.ib(converter=_convert_rhs)
    prec = attr.ib(default=None)
    rassoc = attr.ib(default=False)
    mode = attr.ib(default=0)  # Parser.NORMAL_MODE
    pointer = attr.ib(default=None)
    parent = attr.ib(default=None)

    def with_pointer(self, pointer):
        return attr.evolve(self, pointer=pointer)

    @property
    def next_symbols(self):
        if self.pointer is None:
            raise Exception('Rule has no pointer')
        elif self.pointer == len(self.rhs):
            return frozenset()
        return self.rhs[self.pointer]

    @property
    def advanced(self):
        if self.pointer is None:
            raise Exception('Rule has no pointer')
        elif self.pointer == len(self.rhs):
            raise Exception('Rule cannot be advanced')
        return self.with_pointer(self.pointer + 1)

    def __repr__(self):
        syms = []
        for sym in self.rhs:
            if len(sym) == 1:
                syms.append(str(next(iter(sym))))
            else:
                alts = ' | '.join(str(s) for s in sym)
                syms.append(f'({alts})')
        if self.pointer is not None:
            syms.insert(self.pointer, '.')
        rhs = ' '.join(syms)
        if self.prec is not None:
            return f'{self.lhs} -> {rhs} ({self.prec})'
        return f'{self.lhs} -> {rhs}'


class ItemSet:
    def __init__(self, grammar, items):
        self.items = set(items)

        # complete the itemset by repeatedly finding all of the productions
        # which come after pointers in the set, and adding all the rules that
        # produce them recursively.
        while True:
            old_size = len(self.items)
            nexts = set(
                s for s in chain.from_iterable(
                    r.next_symbols for r in self.items)
                if not s.is_terminal)
            self.items.update(
                grammar.rules[r].with_pointer(0)
                for r in grammar.find_rule_indices(nexts))
            if len(self.items) == old_size:
                break

    @property
    def next_symbols(self):
        """Gets a list of possible next symbols."""
        return set(chain.from_iterable(r.next_symbols for r in self.items))

    def advance(self, symbol):
        """Advances the itemset by the given symbol.

        Returns a frozenset of items where items which can be advanced by the
        given symbol have been, and items which cannot have been dropped.
        """
        return frozenset(
            item.advanced for item in self.items
            if symbol in item.next_symbols)

    @property
    def mode(self):
        """Gets the lexer mode for this itemset."""

        # A simple example of rules using lexer modes:
        #
        #     R -> n S z  (mode U)
        #     S -> x y    (mode V)
        #
        # where n is a mode-0 token and x, y, z are mode U tokens. In this
        # case, we want to accept n, shift to mode V, accept tokens x, y, z,
        # then shift back to mode U. The itemsets for the above rules are:
        #
        #  0. R -> . n S z
        #
        #  1. R -> n . S z
        #   + S -> . x y
        #
        #  2. S -> x . y
        #
        #  3. S -> x y .
        #
        #  4. R -> n S . z
        #
        #  5. R -> n S z .
        #
        # Clearly, in set 0 we want mode U so we can get n, shifting us to set
        # 1. The set stack is now [0, 1], and we want to be in mode V so we can
        # get x for our lookahead.
        #
        # Now we shift set 2, making our set stack [0, 1, 2]. We still want
        # mode V to get y. This lets us shift set 3, set stack [0, 1, 2, 3]. We
        # still want mode V, so we can get z.
        #
        # At this point, we reduce by S, which makes our set stack [0, 1, 4].
        # We don't get a new lookahead after a reduce, so we don't care about
        # the mode.
        #
        # Then we shift by 5, set stack [0, 1, 4, 5]. At this point, we need to
        # be back in mode U for whatever comes after R.
        #
        # From the above we can derive the following set-mode associations.
        # Pairs marked (I) indicates that the mode has the same mode as the
        # previous one on the stack (ignoring set 4, which doesn't really have
        # a mode):  0=U, 1=V, 2=V (I), 3=V (I), 5=U
        #
        # In sets 0 and 1, we're at position 0 in a rule with the appropriate
        # mode, so we assign the mode based on that. Set 2 can inherit. In set
        # 5, we're at the end of a rule with the appropriate mode. Assigning
        # based on that also assigns mode V to set 3, which is not necessary,
        # but harmless.
        #
        # Therefore, we only care about rules where the pointer is at the
        # beginning or end.
        modes = {r.mode for r in self.items
                 if r.pointer == 0 or len(r.next_symbols) == 0}

        # If there are no such rules, we can inherit. Note that we can't assign
        # based on the rule that we're in the middle of, because then we'd be
        # trying to assign mode U to set 1, which we're already assigning mode
        # V to.
        if len(modes) == 0:
            return Parser.INHERIT_MODE

        # If we end up with two rules giving us conflicting modes, we will
        # consider that an error in the grammar.
        assert len(modes) == 1, f"mode/mode conflict: {self.items}"
        return modes.pop()


@attr.s(slots=True, frozen=True, repr=False, cmp=False)
class Special:
    name = attr.ib()

    def __repr__(self):
        return "$" + self.name


class Grammar:
    EMPTY_TOKEN = Special("EMPTY")
    EMPTY = Symbol(EMPTY_TOKEN)
    END = Special("END")

    def __init__(self, start_symbol, end_symbols, rules):
        self.rules = rules
        self.start_symbol = _convert_lhs(start_symbol)
        self.end_symbols = frozenset(_convert_lhs(s) for s in end_symbols)

    @property
    def symbols(self):
        ts = set()
        for rule in self.rules:
            ts.add(rule.lhs)
            for syms in rule.rhs:
                ts.update(syms)
        return ts

    def find_rule_indices(self, symbols):
        """Returns a list of rule indices which produce the given symbols."""
        return [k for k, r in enumerate(self.rules) if r.lhs in symbols]

    def find_starting_rule_index(self):
        """Finds the starting rule given the start symbol."""
        starts = self.find_rule_indices([self.start_symbol])
        if len(starts) == 0:
            raise Exception('No starting rule found')
        elif len(starts) > 1:
            raise Exception('Multiple starting rules found')
        elif len(self.rules[starts[0]].rhs) != 1:
            raise Exception('Starting rule must have one token')
        else:
            return starts[0]

    def build_firstsets(self):
        """Builds the First sets for every extended symbol.

        The First set is the set of all terminals which can grammatically
        appear at the beginning of a given symbol.
        """

        start_time = time.perf_counter()
        firstsets = {}

        # pre-populate with empty firstsets (for nonterminals) and identity
        # firstsets (for terminals)
        for sym in self.symbols:
            if sym.is_terminal:
                firstsets[sym] = {sym.parent}
            else:
                firstsets[sym] = set()

        # 1. if V -> x, then First(V) contains x
        # 2. if V -> (), then First(V) contains ()
        nzrules = []
        for rule in self.rules:
            if len(rule.rhs) == 0:
                firstsets[rule.lhs].add(self.EMPTY)
                continue
            for sym in rule.rhs[0]:
                if sym.is_terminal:
                    firstsets[rule.lhs].update(firstsets[sym])
                else:
                    # cache rules that rule 3 applies to in advance
                    nzrules.append(rule)

        # 3. if V -> A B C, then First(V) contains First(A) - (). If First(A)
        #    contains (), then First(V) also contains First(B), etc. If A, B,
        #    and C all contain (), then First(V) contains ().
        #
        # Since rules can be (mutually) left-recursive, we may have to apply
        # this rule multiple times to catch everything.
        updated = True
        count = 0
        while updated:
            count += 1
            updated = False
            for rule in nzrules:
                # we know in advance that these rules begin with a nonterminal
                # on the right-hand side, because they're the ones we cached
                # when applied rules 1 & 2.
                old_len = len(firstsets[rule.lhs])
                for symbols in rule.rhs:
                    fs = frozenset(chain.from_iterable(
                        firstsets[s] for s in symbols))
                    if self.EMPTY not in fs:
                        firstsets[rule.lhs].update(fs)
                        break
                    firstsets[rule.lhs].update(fs - {self.EMPTY})
                else:
                    firstsets[rule.lhs].add(self.EMPTY)
                if len(firstsets[rule.lhs]) > old_len:
                    updated = True

        end_time = time.perf_counter()
        elapsed_ms = (end_time - start_time) * 1000
        logger.debug(__(
            "Built firstsets ({} cycles) in {:.2f}ms",
            count, elapsed_ms))
        return firstsets

    def build_followsets(self):
        """Builds the Follow sets for every extended symbol.

        The Follow set is the set of all terminals which can grammatically
        appear after the given symbol.
        """

        firstsets = self.build_firstsets()
        start_time = time.perf_counter()
        followsets = {}

        # pre-populate with empty followsets
        for sym in self.symbols:
            if sym == self.start_symbol:
                followsets[sym] = {s.parent for s in self.end_symbols}
            else:
                followsets[sym] = set()

        # suppose we have a rule R -> a*Db. Then we add First(b) to Follow(D).
        for rule in self.rules:
            for k in range(len(rule.rhs) - 1):
                for s in rule.rhs[k]:
                    if not s.is_terminal:
                        for t in rule.rhs[k+1]:
                            followsets[s].update(firstsets[t])

        # suppose we have a rule R -> a*D. Then we add Follow(R) to Follow(D).
        # Because we can end up with irritating things like two follow sets
        # mutually depending on each other, we've handled this by just applying
        # the rule until we reach a fixed state.
        updated = True
        count = 0
        while updated:
            count += 1
            updated = False
            for rule in self.rules:

                # empty rules don't tell us anything for this pass
                if len(rule.rhs) == 0:
                    continue

                for s in rule.rhs[-1]:
                    if not s.is_terminal:
                        old_len = len(followsets[s])
                        followsets[s].update(
                            followsets[rule.lhs])
                        if len(followsets[s]) > old_len:
                            updated = True

        end_time = time.perf_counter()
        elapsed_ms = (end_time - start_time) * 1000
        logger.debug(__(
            "Build followsets ({} cycles) in {:.2f}ms",
            count, elapsed_ms))
        return followsets

    def build_parser(self, hidden=None, channel=ReStream.CHANNEL_DEFAULT):
        logger.debug(__("Grammar has {} rules", len(self.rules)))

        start_time_t = time.perf_counter()
        translation_table = TranslationTable(self)
        extended_grammar = translation_table.build_extended_grammar()
        modes = translation_table.build_modes()
        followsets = extended_grammar.build_followsets()
        start_time = time.perf_counter()

        # Build the action/goto table. This is what the parse function actually
        # uses. In the action part of the table (where the input is a
        # terminal), there are two possible actions: shift, and reduce. These
        # are represented by an (action, index) tuple. In response to a reduce,
        # the parser will execute a goto by providing a nonterminal as input.
        # These are represented as integers.
        agtable = {}
        conflicts = []

        # copy the nonterminal entries in the translation table over as gotos
        # and the terminal entries as shifts.
        for (f, s), t in translation_table.items():
            if s.is_terminal:
                agtable[(f, s.value)] = ('shift', None, t)
            else:
                agtable[(f, s.value)] = t  # goto

        # construct the final sets by merging extended rules which are based on
        # the same rule and have the same end point.
        finalset_rules = [None] * len(translation_table.itemsets)
        finalset_followsets = [set() for _ in translation_table.itemsets]
        for rule in extended_grammar.rules:
            if len(rule.rhs) == 0:
                # if the rule has no rhs, then the starting point is the same
                # as the ending point.
                final = rule.lhs.start
            else:
                finals = {s.end for s in rule.rhs[-1]}
                assert len(finals) == 1
                final = finals.pop()
            if finalset_rules[final] is not None \
               and finalset_rules[final] != rule.parent:
                # This is a reduce/reduce conflict. We decide how to resolve
                # this based on the precedence of the rules involved.
                # TODO work out how to handle ties?
                # TODO check if this is sound. Somehow
                if finalset_rules[final].prec is None or \
                   rule.parent.prec is None or \
                   finalset_rules[final].prec == rule.parent.prec:
                    conflicts.append(
                        f'reduce/reduce:\n'
                        f'  {finalset_rules[final]}\n'
                        f'  {rule.parent}')
                    continue

                logging.debug(__(
                    "Resolved reduce/reduce conflict between {} and {}",
                    finalset_rules[final],
                    rule.parent))
                if finalset_rules[final].prec > rule.parent.prec:
                    continue

            finalset_rules[final] = rule.parent
            finalset_followsets[final].update(followsets[rule.lhs])

        # add the merged reductions to the table
        for k, followset in enumerate(finalset_followsets):
            for symbol in followset:
                if (k, symbol.value) in agtable:
                    # This is a shift/reduce conflict. We decide how to resolve
                    # this based on the precedence of the rules involved.

                    # Note that the shift index is a state number, not a rule
                    # number. State numbers correspond to item sets. In
                    # particular, we're looking for the rule that has already
                    # been partially applied.
                    _, _, shift_index = agtable[(k, symbol.value)]
                    partials = [
                        i for i
                        in translation_table.itemsets[shift_index].items
                        if i.pointer > 0]
                    assert len(partials) == 1, 'shift/reduce (GENERATOR BUG)'
                    shift_rule = partials[0]

                    # If one of them is missing a precedence, go ahead and
                    # hard-fail.
                    if shift_rule.prec is None \
                       or finalset_rules[k].prec is None:
                        conflicts.append(
                            f'shift/reduce:\n'
                            f'  {shift_rule}\n'
                            f'  {finalset_rules[k]}')
                        continue

                    # If the shifting rule is right-associative, then we should
                    # break ties in favour of the shift. Otherwise, in favour
                    # of the reduce. Note because we look at the shift rule's
                    # associativity for this decision, a right-associative rule
                    # will bind more tightly than a left-associative rule with
                    # the same precedence.
                    if (shift_rule.prec > finalset_rules[k].prec
                        or (shift_rule.rassoc
                            and shift_rule.prec == finalset_rules[k].prec)):
                        continue

                # Check to see if this is actually the accept state
                if finalset_rules[k].lhs == self.start_symbol \
                   and symbol in self.end_symbols:
                    agtable[(k, symbol.value)] = ('accept', None, None)
                else:
                    agtable[(k, symbol.value)] = (
                        'reduce',
                        finalset_rules[k].lhs.value,
                        len(finalset_rules[k].rhs))

        if len(conflicts) > 0:
            logger.critical('\n\n'.join(conflicts))
            assert False, "conflicts detected"

        # build the hidden-channel parsers
        hidden_parsers = {
            channel: aux_grammar.build_parser(channel=channel)
            for channel, aux_grammar in (hidden or {}).items()
        }

        parser = Parser(agtable, modes, hidden_parsers, channel)
        end_time = time.perf_counter()
        elapsed_ms = (end_time - start_time) * 1000
        elapsed_t_ms = (end_time - start_time_t) * 1000
        logger.debug(__(
            "Built action/goto table ({} enties) in {:.2f}ms",
            len(agtable), elapsed_ms))
        logger.debug(__("Built parser in {:.2f}ms total", elapsed_t_ms))
        logger.debug(__(
            "Symbols: {}, States: {}",
            len(self.symbols), len(translation_table.itemsets)))
        return parser


class TranslationTable:
    """A table of itemset/state transitions."""

    def __init__(self, grammar):
        self.end_symbols = grammar.end_symbols
        self.translation_table = {}
        self.itemsets = []
        self.itemset_index = {}

        start_time = time.perf_counter()

        # The first item set is based around the start rule
        start = grammar.find_starting_rule_index()
        startitem = grammar.rules[start].with_pointer(0)
        self.itemsets.append(ItemSet(grammar, {startitem}))
        self.itemset_index[frozenset({startitem})] = 0

        # next, we work our way down the itemsets, advancing them using the
        # allowed productions. The resulting items are used to construct new
        # itemsets. We also build the translation table as we go
        current = 0
        while current < len(self.itemsets):
            for symbol in self.itemsets[current].next_symbols:
                key = self.itemsets[current].advance(symbol)
                if key in self.itemset_index:
                    itemset = self.itemset_index[key]
                else:
                    itemset = len(self.itemsets)
                    self.itemset_index[key] = itemset
                    self.itemsets.append(ItemSet(grammar, key))
                self.translation_table[(current, symbol)] = itemset
            current += 1

        end_time = time.perf_counter()
        elapsed_ms = (end_time - start_time) * 1000
        logger.debug(__("Built {} itemsets in {:.2f}ms", current, elapsed_ms))

    def items(self):
        return self.translation_table.items()

    def build_extended_grammar(self):
        """Builds the extended rule set.

        This produces a set of rules where each symbol sym has been replaced
        with the triple (s0, sym, s1) where s0 is the state/itemset preceding
        that symbol, and s1 is the state/itemset following it. If the rule can
        be applied from multiple states (i.e. it shows up in multiple
        itemsets), it will show up multiple times with different state numbers.
        """
        start_time = time.perf_counter()
        extended_rules = set()

        for current, itemset in enumerate(self.itemsets):
            rules_to_process = list(itemset.items)
            while len(rules_to_process) > 0:
                rule = rules_to_process.pop()
                if rule.pointer != 0:
                    continue
                prev = None
                state = current
                rhs = []
                abort = False
                parent = rule.parent
                if parent is None:
                    parent = rule
                for k, symbols in enumerate(rule.rhs):
                    prev = state
                    states = {self.translation_table[(state, s)]
                              for s in symbols}
                    if len(states) == 1:
                        state = states.pop()
                        rhs.append(tuple(
                            s.extend(prev, state) for s in symbols))
                        continue

                    # our alternation goes to different places, so split
                    # the rule, and try again.
                    for s in symbols:
                        rhs = list(rule.rhs)
                        rhs[k] = s
                        rules_to_process.append(
                            attr.evolve(rule, rhs=rhs, parent=rule))
                        abort = True
                    break
                if abort:
                    continue
                try:
                    lhs = rule.lhs.extend(
                        current, self.translation_table[(current, rule.lhs)])
                except KeyError:
                    lhs = rule.lhs.extend(current, Grammar.END)
                    start_symbol = lhs
                extended_rules.add(
                    attr.evolve(rule, lhs=lhs, rhs=rhs, pointer=None,
                                parent=parent.with_pointer(None)))

        extended_grammar = Grammar(
            start_symbol,
            [s.extend(None, None) for s in self.end_symbols],
            extended_rules)
        end_time = time.perf_counter()
        elapsed_ms = (end_time - start_time) * 1000
        sz = len(extended_rules)
        logger.debug(__(
            "Built extended grammar ({} rules) in {:.2f}ms",
            sz, elapsed_ms))
        return extended_grammar

    def build_modes(self):
        """Builds the lexer mode table.

        Depending on what rule we are currently following, we ask the lexer to
        operate in different modes. Each itemset/state is associated with one
        mode, that of the rule(s) which are currently in-progress.
        """

        return [s.mode for s in self.itemsets]


class Parser:
    NORMAL_MODE = 0
    INHERIT_MODE = -1

    def __init__(self, agtable, modes, hidden, channel):
        self.agtable = agtable
        self.modes = modes
        self.hidden = hidden
        self.channel = channel

    def select_mode(self, set_stack):
        return next((self.modes[s]
                     for s in reversed(set_stack)
                     if self.modes[s] != self.INHERIT_MODE),
                    self.NORMAL_MODE)

    def next_token_skip_hidden(self, stream, next_token, set_stack):
        while True:
            lookahead = next_token(stream, self.select_mode(set_stack))
            if lookahead.channel == self.channel \
               or lookahead.channel == ReStream.CHANNEL_ALL:
                return lookahead

            # When a token comes in on a channel other than the one we're
            # handling, we delegate to another parser for that channel, which
            # consumes the input. This is useful for things like comments,
            # which can show up anywhere -- handling them in the main grammar
            # would be impossible.
            stream.rewind(lookahead)
            p = self.hidden[lookahead.channel]
            p(stream, next_token, lambda t, s, c, m: None)

    def __call__(self, stream, next_token, make_node):
        """Parses a given input.

        This method may be called multiple times and does not modify the
        object.

        'next_token' must be a callable which takes two arguments: the
        'stream', and an int for the mode, which is 0 initially. It must return
        a Token.

        'make_node' must be a callable, which is called every time a reduction
        is performed. It is passed three arguments: the nonterminal being
        reduced, a span covering the tokens involved in the reduction, and an
        iterable of the children of the reduction, which are a mix of Tokens
        and values returned from make_node.
        """

        start_time = time.perf_counter()
        output = []
        set_stack = [0]
        lookahead = self.next_token_skip_hidden(stream, next_token, set_stack)

        while True:
            try:
                action, sym, arg = self.agtable[(set_stack[-1], lookahead.t)]
            except KeyError:
                try:
                    action, sym, arg = self.agtable[
                        (set_stack[-1], Grammar.EMPTY_TOKEN)]
                    assert action == 'reduce'
                except KeyError:
                    msg = [f"Got {lookahead.t} but expected one of:"]
                    for state, token in self.agtable:
                        if state == set_stack[-1]:
                            msg.append(f"  {token}")
                    raise ParseError("\n".join(msg))

            if action == 'shift':
                output.append((lookahead, lookahead.span))
                set_stack.append(arg)
                lookahead = self.next_token_skip_hidden(
                    stream, next_token, set_stack)
            elif action == 'reduce':
                if arg > 0:
                    children, spans = zip(*output[-arg:])
                    span = TextSpan.cover(spans)
                    del output[-arg:]
                    del set_stack[-arg:]
                else:
                    children = []
                    # Since we are reducing an empty rule, we know by [vigorous
                    # handwaving] that the lack-of-tokens we're trying to
                    # reduce is bounded on the left by the last item in the
                    # output stack (if present) and on the right by the
                    # lookahead token. Note that this approach may result in
                    # the span corresponding to a block of whitespace or a
                    # comment.
                    end = lookahead.span.start
                    if len(output) > 0:
                        start = output[-1][1].end
                    else:
                        start = end
                    span = TextSpan(*start, *end)
                set_stack.append(self.agtable[(set_stack[-1], sym)])
                output.append((make_node(sym, span, children,
                                         self.modes[set_stack[-1]]),
                               span))
            else:
                assert action == 'accept'
                break

        assert len(output) == 1

        end_time = time.perf_counter()
        elapsed_ms = (end_time - start_time) * 1000
        if self.channel == ReStream.CHANNEL_DEFAULT:
            logger.debug(__(
                "Parsed input on channel {} in {:.2f}ms",
                self.channel, elapsed_ms))
        return output[0][0]



# jeff65 linker image manager
# Copyright (C) 2017  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
import struct
from . import symbol, types


def make_startup_for(main, version):
    archive = symbol.Archive()
    archive.symbols['$startup.__start'] = symbol.Symbol(
        section='startup',
        data=struct.pack(
            # Defines a simple startup header. Note that in theory, we could
            # simply make the target of the SYS instruction be our main
            # function, but our relocation system isn't smart enough for that
            # and we may wish to add more complicated startup code in the
            # future. We could assume that main exits using RTS and save one
            # byte and a few cycles, but for now it's convenient to have a
            # well-known address to break on.
            "<HHB4sxHBHB",
            0x080b,         # 0x0000 0x0801 (H)    addr of next BASIC line
            version,        # 0x0002 0x0803 (H)    BASIC line number
            0x9e, b'2061',  # 0x0004 0x0805 (B4sx) SYS2061 (0x080d)
            0x0000,         # 0x000a 0x080b (H)    BASIC end-of-program
            0x20, 0xffff,   # 0x000c 0x080d (BH)   jsr $ffff
            0x60,           # 0x000f 0x0810 (B)    rts
        ),
        type_info=types.phantom,
        relocations={
            # Relocation for the address of 'main'
            0x000d: symbol.Relocation(main),
        })
    return archive


class Image:
    m_reloc = re.compile(r'^([^,+-]+)([+-]\d+)?(,lo|,hi)?$')

    def __init__(self, fileobj, base_address=0x0801):
        self.fileobj = fileobj
        self.offsets = {}
        self.archive = symbol.Archive()
        self.base_address = base_address
        self.start_symbol = '$startup.__start'

        # Header for PRG files. Identifies the load location in memory.
        # 0x0801 is the load location for BASIC programs.
        self.prg_header = struct.pack("<H", self.base_address)

    def add_archive(self, archive):
        self.archive.update(archive)

    def current_offset(self):
        return self.fileobj.tell() - len(self.prg_header)

    def seek_to_offset(self, name, inc):
        self.fileobj.seek(self.offsets[name] + inc + len(self.prg_header))

    def _emit_sym(self, name, sym):
        self.offsets[name] = self.current_offset()
        self.fileobj.write(sym.data)

    def link(self):
        """Links the image, writing it out to a file."""
        # write out the prg header
        self.fileobj.write(self.prg_header)

        # write out the startup section
        startup = self.archive.find_section('startup')
        assert len(startup) == 1
        for name, sym in startup:
            self._emit_sym(name, sym)

        # write out the text section
        for name, sym in self.archive.find_section('text'):
            self._emit_sym(name, sym)

        # perform relocations
        for name, offset, reloc in self.archive.relocations():
            if name not in self.offsets:
                continue

            self.seek_to_offset(name, offset)
            addr = reloc.bind(name).compute_bin(
                self.base_address, self.offsets)
            self.fileobj.write(addr)



# jeff65 linker symbol data structures
# Copyright (C) 2017  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import collections
import collections.abc
import io
import mmap
import struct
import warnings
import zlib
from . import types
from .fmt import Fmt


MAX_STRING_SIZE = (1 << 31) - 1
MAX_BLOB_SIZE = (1 << 31) - 1
ARCHIVE_MAGIC = b'\x93Blm\x0d\x0a\x1a\x0a'


class Archive:
    """An archive is a collection of compiled symbols.

    The compiler produces an archive for each unit, then the linker combines
    archives into a program image.
    """

    def __init__(self, fileobj=None):
        self.symbols = {}
        if fileobj:
            self.load(fileobj)

    def update(self, archive):
        """Updates this archive with contents of another archive.

        Items already present in this archive are retained, and items present
        in both archives are replaced with the item from the other archive.
        """
        self.symbols.update(archive.symbols)

    def load(self, fileobj):
        with ArchiveReader(fileobj, closefd=False) as reader:
            self.update(reader.load())

    def dump(self, fileobj):
        with ArchiveWriter(fileobj, closefd=False) as writer:
            writer.dump(self)

    def dumpf(self, path):
        with open(path, 'wb') as f:
            self.dump(f)

    def loadf(self, path):
        with open(path, 'rb') as f:
            self.load(f)

    def find_section(self, section):
        return [(name, sym)
                for name, sym in self.symbols.items()
                if sym.section == section]

    def relocations(self):
        for name, sym in self.symbols.items():
            for offset, reloc in sym.relocations.items():
                yield (name, offset, reloc)


class Relocation:
    full = ord('w')
    hi = ord('h')
    lo = ord('l')

    fields = [
        ('symbol', Fmt.str, Fmt.make_cc('sy'), True),
        ('increment', Fmt.i16, Fmt.make_cc('ic'), True),
        ('byte', Fmt.u8, Fmt.make_cc('by'), True),
    ]

    def __init__(self, symbol, increment=0, byte=None):
        self.symbol = symbol
        self.increment = increment
        self.byte = byte or self.full

    def __eq__(self, other):
        return (isinstance(other, Relocation)
                and self.symbol == other.symbol
                and self.increment == other.increment
                and self.byte == other.byte)

    def __repr__(self):
        return 'Relocation({}, increment={}, byte={})'.format(
            self.symbol, self.increment, repr(self.byte))

    def bind(self, symbol):
        if self.symbol is None:
            return Relocation(symbol, self.increment, self.byte)
        return self

    def compute_offset(self, base, offsets):
        return base + offsets[self.symbol] + self.increment

    def compute_value(self, base, offsets):
        offset = self.compute_offset(base, offsets)
        if self.byte == self.lo:
            return offset & 0x00ff
        elif self.byte == self.hi:
            return offset >> 8
        return offset

    def compute_bin(self, base, offsets):
        offset = self.compute_offset(base, offsets)
        bin = struct.pack('<H', offset)
        if self.byte == self.lo:
            return bin[0:1]
        elif self.byte == self.hi:
            return bin[1:2]
        return bin

    def validate(self):
        assert isinstance(self.symbol, str)
        assert isinstance(self.increment, int)
        assert self.byte in [self.full, self.hi, self.lo]

    @classmethod
    def _empty(cls):
        return cls(None)


class Symbol:
    discriminator = Fmt.make_cc('Sy')
    _fmt_relocation_table = Fmt.table(Fmt.u16, Fmt.struct(Relocation))
    fields = [
        ('section', Fmt.str, Fmt.make_cc('sc'), True),
        ('type_info', types.fmt_type_info, Fmt.make_cc('ty'), True),
        ('relocations', _fmt_relocation_table, Fmt.make_cc('re'), True),
        ('data', Fmt.blob, Fmt.make_cc('da'), True),
    ]

    def __init__(self, section, data, type_info, relocations=None):
        self.section = section
        self.data = data
        self.type_info = type_info
        if relocations is not None:
            self.relocations = collections.OrderedDict(relocations)
        else:
            self.relocations = collections.OrderedDict()

    def __eq__(self, other):
        return (isinstance(other, Symbol)
                and self.section == other.section
                and self.data == other.data
                and self.type_info == other.type_info
                and self.relocations == other.relocations)

    def validate(self):
        assert isinstance(self.section, str)
        assert isinstance(self.data, bytes)
        assert self.type_info is not None
        assert isinstance(self.relocations, collections.abc.Mapping)

    @classmethod
    def _empty(cls):
        return cls(None, None, None)


class ArchiveError(Exception):
    pass


class ArchiveWriter:
    def __init__(self, fileobj, closefd=True):
        self.fileobj = fileobj
        self.closefd = closefd
        self.handlers = {
            Fmt._bool: self.dump_fmt('?'),
            Fmt._u8: self.dump_fmt('B'),
            Fmt._u16: self.dump_fmt('H'),
            Fmt._u32: self.dump_fmt('L'),
            Fmt._i8: self.dump_fmt('b'),
            Fmt._i16: self.dump_fmt('h'),
            Fmt._i32: self.dump_fmt('l'),
            Fmt._str: self.dump_string,
            Fmt._blob: self.dump_blob,
            Fmt._array: self.dump_array,
            Fmt._table: self.dump_table,
            Fmt._struct: self.dump_struct,
            Fmt._union: self.dump_union,
        }
        self.blob_infos = collections.deque()
        self.crc = 0
        self._depth = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def close(self):
        if self.closefd:
            self.fileobj.close()

    def _write(self, *datas):
        offset = self.fileobj.tell()
        for data in datas:
            assert len(data) == self.fileobj.write(data)
            self.crc = zlib.crc32(data, self.crc)
        return offset

    def dump(self, archive):
        # write out a placeholder header.
        self.fileobj.seek(0)
        self.dump_header(0, 0, 0, placeholder=True)

        entry_offset, entry_len, entry_crc = 0, 0, 0
        for name, symbol in archive.symbols.items():
            # first get the blobs out in traversal order, and write them out.
            # Keep track of the offsets and checksums.
            for blob in Fmt.find_blobs(symbol):
                self.crc = 0
                blob_offset = self._write(blob)
                self.blob_infos.append((blob_offset, len(blob), self.crc))

            # next we emit the structure.
            self.crc = 0
            sym_offset = self.dump_union([Symbol], symbol)
            sym_len = self.fileobj.tell() - sym_offset
            sym_crc = self.crc
            assert len(self.blob_infos) == 0  # should have consumed all blobs

            # finally, the entry
            self.crc = 0
            entry_offset = self.dump_entry(
                sym_offset, sym_len, sym_crc,
                entry_offset, entry_len, entry_crc, name)
            entry_len = self.fileobj.tell() - entry_offset
            entry_crc = self.crc

        # replace the header with the real one
        self.fileobj.seek(0)
        self.dump_header(entry_offset, entry_len, entry_crc)

    def dump_header(self, next_off, next_len, next_crc, placeholder=False):
        if placeholder:
            magic = b'\xff' * len(ARCHIVE_MAGIC)
        else:
            magic = ARCHIVE_MAGIC
        return self._write(struct.pack('<8s3L', magic,
                                       next_off, next_len, next_crc))

    def dump_entry(self,
                   data_off, data_len, data_crc,
                   next_off, next_len, next_crc,
                   name):
        enc_name = name.encode('utf8')
        fixed_part = struct.pack('<6Ll',
                                 data_off, data_len, data_crc,
                                 next_off, next_len, next_crc,
                                 len(enc_name))
        return self._write(fixed_part, enc_name)

    def dump_by_type(self, t, obj):
        return self.handlers[t[0]](*t[1:], obj)

    def dump_string(self, data):
        enc_data = data.encode('utf8')
        sz = struct.pack('<l', len(enc_data))
        return self._write(sz, enc_data)

    def dump_fmt(self, fmt):
        return lambda data: self._write(struct.pack('<' + fmt, data))

    def dump_struct(self, ty, obj):
        assert type(obj) == ty
        field_count = len([None for _, _, _, pack in obj.fields if pack])
        offset = self._write(struct.pack('<H', field_count))
        for field, t, cc, pack in obj.fields:
            if not pack:
                continue
            self._write(struct.pack('<H', cc))
            self.dump_by_type(t, getattr(obj, field))
        return offset

    def dump_union(self, tys, obj):
        assert type(obj) in tys
        offset = self._write(struct.pack('<H', obj.discriminator))
        self.dump_struct(type(obj), obj)
        return offset

    def dump_array(self, t, objs):
        offset = self._write(struct.pack('<L', len(objs)))
        for obj in objs:
            self.dump_by_type(t, obj)
        return offset

    def dump_table(self, tkey, tvalue, table):
        offset = self._write(struct.pack('<L', len(table)))
        for key, value in table.items():
            self.dump_by_type(tkey, key)
            self.dump_by_type(tvalue, value)
        return offset

    def dump_blob(self, obj):
        return self._write(struct.pack('<LLL', *self.blob_infos.popleft()))


class ArchiveReader:
    def __init__(self, fileobj, closefd=True):
        self.fileobj = fileobj
        self.closefd = closefd
        if isinstance(fileobj, io.BytesIO):
            self._mapping = None
            self.mmap = fileobj.getbuffer()
        else:
            self._mapping = mmap.mmap(self.fileobj.fileno(), 0,
                                      access=mmap.ACCESS_READ)
            self.mmap = memoryview(self._mapping)
        self.handlers = {
            Fmt._bool: self.load_fmt('?'),
            Fmt._u8: self.load_fmt('B'),
            Fmt._u16: self.load_fmt('H'),
            Fmt._u32: self.load_fmt('L'),
            Fmt._i8: self.load_fmt('b'),
            Fmt._i16: self.load_fmt('h'),
            Fmt._i32: self.load_fmt('l'),
            Fmt._str: self.load_string,
            Fmt._blob: self.load_blob,
            Fmt._array: self.load_array,
            Fmt._table: self.load_table,
            Fmt._struct: self.load_struct,
            Fmt._union: self.load_union,
        }

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def close(self):
        self.mmap.release()
        self.mmap = None
        if self._mapping is not None:
            self._mapping.close()
            self._mapping = None
        if self.closefd:
            self.fileobj.close()

    def load(self):
        archive = Archive()
        entry_off, entry_len, entry_crc = self.load_header()

        while entry_off != 0:
            # load the next entry. This method does its own CRC checking
            (sym_off, sym_len, sym_crc,
             entry_off, entry_len, entry_crc,
             name) = self.load_entry(entry_off, entry_len, entry_crc)

            # load the named symbol
            assert sym_crc == zlib.crc32(self.mmap[sym_off:sym_off+sym_len])
            sym_end, symbol = self.load_union([Symbol], sym_off)
            assert sym_end == sym_off+sym_len
            archive.symbols[name] = symbol

        return archive

    def load_header(self):
        magic, entry_off, entry_len, entry_crc = struct.unpack_from(
            '<8s3L', self.mmap, 0)
        if magic != ARCHIVE_MAGIC:
            # Header did not match
            # TODO attempt to diagnose.
            raise ArchiveError("File is not a valid blum archive")
        return entry_off, entry_len, entry_crc

    def load_entry(self, offset, sz, crc):
        assert crc == zlib.crc32(self.mmap[offset:offset+sz])
        fmt = '<6Ll'
        (data_off, data_len, data_crc,
         next_off, next_len, next_crc,
         name_len) = struct.unpack_from(fmt, self.mmap, offset)
        name_start = offset + struct.calcsize(fmt)
        name_end = name_start + name_len
        assert name_end <= offset+sz
        with self.mmap[name_start:name_end] as bview:
            name = bytes(bview).decode('utf8')
        return data_off, data_len, data_crc, next_off, next_len, next_crc, name

    def load_by_type(self, t, off):
        return self.handlers[t[0]](*t[1:], off)

    def load_string(self, off):
        off, sz = self.load_fmt('l', off)
        compressed = sz < 0
        end = off + abs(sz)

        with self.mmap[off:end] as bview:
            if compressed:
                dc = zlib.decompressobj()
                bdata = dc.decompress(bview, max_length=MAX_STRING_SIZE)
                if len(dc.unconsumed_tail) > 0:
                    warnings.warn("Truncated large (>2 GiB) string in archive")
            else:
                bdata = bytes(bview)
            return end, bdata.decode('utf8')

    def load_fmt(self, fmt, *args):
        fmt = '<' + fmt
        sz = struct.calcsize(fmt)

        def load_fmt_inner(off):
            val, = struct.unpack_from(fmt, self.mmap, off)
            return off+sz, val
        if len(args) > 0:
            return load_fmt_inner(*args)
        return load_fmt_inner

    def load_struct(self, ty, *args):
        def load_struct_inner(off):
            field_map = {cc: (field, ft) for field, ft, cc, _ in ty.fields}
            off, field_count = self.load_fmt('H', off)
            obj = ty._empty()
            for _ in range(field_count):
                off, cc = self.load_fmt('H', off)
                field, ft = field_map[cc]
                off, val = self.load_by_type(ft, off)
                try:
                    setattr(obj, field, val)
                except AttributeError as ex:
                    raise ArchiveError(
                        "Can't set {} on {}".format(field, type(obj))) from ex
            obj.validate()
            return off, obj
        if len(args) > 0:
            return load_struct_inner(*args)
        return load_struct_inner

    def load_union(self, types, *args):
        def load_union_inner(off):
            off, disc = self.load_fmt('H', off)
            sel = next(t for t in types if t.discriminator == disc)
            return self.load_struct(sel, off)
        if len(args) > 0:
            return load_union_inner(*args)
        return load_union_inner

    def load_array(self, t, off):
        objs = []
        off, sz = self.load_fmt('L', off)
        for _ in range(sz):
            off, val = self.load_by_type(t, off)
            objs.append(val)
        return off, objs

    def load_table(self, tkey, tvalue, off):
        table = collections.OrderedDict()
        off, sz = self.load_fmt('L', off)
        for _ in range(sz):
            off, key = self.load_by_type(tkey, off)
            off, val = self.load_by_type(tvalue, off)
            table[key] = val
        return off, table

    def load_blob(self, off):
        fmt = '<LLL'
        blob_off, blob_len, blob_crc = struct.unpack_from(
            fmt, self.mmap, off)
        off += struct.calcsize(fmt)
        with self.mmap[blob_off:blob_off+blob_len] as bview:
            assert blob_crc == zlib.crc32(bview)
            return off, bytes(bview)



# jeff65 gold-syntax assembler passes
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import struct
from ... import ast, pattern
from ...pattern import Predicate as P


class AssemblyError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def asmrun(fmt, *args):
    return ast.AstNode('asmrun', attrs={
        'bin': struct.pack(fmt, *args)
    })


@pattern.transform(pattern.Order.Any)
class AssembleWithRelocations:
    @pattern.match(
        ast.AstNode('lda', {
            'storage': ast.AstNode('immediate_storage', {
                'value': P('value'),
                'width': P.require(1, AssemblyError),
            }),
        }))
    def lda_imm(self, value):
        return asmrun('<BB', 0xa9, value)

    @pattern.match(
        ast.AstNode('sta', {
            'storage': ast.AstNode('absolute_storage', {
                'address': P('address'),
                'width': P.require(1, AssemblyError),
            })
        }))
    def sta_abs(self, address):
        return asmrun('<BH', 0x8d, address)

    @pattern.match(
        ast.AstNode('jmp', {
            'storage': ast.AstNode('absolute_storage', {
                'address': P('address'),
            })
        }))
    def jmp(self, address):
        return asmrun('<BH', 0x4c, address)

    @pattern.match(ast.AstNode('rts'))
    def rts(self):
        return asmrun('<B', 0x60)


@pattern.transform(pattern.Order.Ascending)
class FlattenSymbol:
    @pattern.match(
        ast.AstNode('block', {
            'stmt': ast.AstNode('asmrun', {'bin': P('bin0')}),
            'next': ast.AstNode('block', {
                'stmt': ast.AstNode('asmrun', {'bin': P('binf')}),
            }),
        }))
    def concatenate_bins(self, bin0, binf):
        return ast.AstNode('block', {
            'stmt': ast.AstNode('asmrun', {'bin': bin0 + binf})
        })

    @pattern.match(
        ast.AstNode('fun', attrs={
            'name': P('name'),
            'type': P('ty'),
            'body': ast.AstNode('block', {
                'stmt': ast.AstNode('asmrun', {'bin': P('text')})
            }),
        }))
    def fun(self, name, ty, text):
        return ast.AstNode('fun_symbol', {
            'name': name,
            'type': ty,
            'text': text,
        })


def lda(storage, span):
    return ast.AstNode('lda', span=span, attrs={
        'size': 2,
        'storage': storage,
    })


def sta(storage, span):
    return ast.AstNode('sta', span=span, attrs={
        'size': 3,
        'storage': storage,
    })


def jmp(storage, span):
    return ast.AstNode('jmp', span=span, attrs={
        'size': 3,
        'storage': storage,
    })


def rts(span):
    return ast.AstNode('rts', span=span, attrs={
        'size': 1,
    })



# jeff65 gold-syntax compiler sequence
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import sys
from . import grammar
from .. import ast, blum, parsing
from .passes import asm, binding, lower, resolve, simplify, typepasses

logger = logging.getLogger(__name__)


passes = [
    resolve.ResolveUnits,
    binding.ShadowNames,
    typepasses.ConstructTypes,
    binding.BindNamesToTypes,
    resolve.ResolveMembers,
    typepasses.PropagateTypes,
    binding.EvaluateConstants,
    binding.ResolveConstants,
    resolve.ResolveStorage,
    lower.LowerAssignment,
    lower.LowerFunctions,
    asm.AssembleWithRelocations,
    asm.FlattenSymbol,
]


def open_unit(unit):
    if str(unit) == "-":
        return sys.stdin
    return open(unit, 'r')


def parse(fileobj, name):
    stream = parsing.ReStream(fileobj)
    tree = grammar.parse(
        stream, grammar.lex,
        lambda t, s, c, m: ast.AstNode(t, span=s, attrs={
            f"{k:02}": v for k, v in enumerate(c)
        }))
    # if parser._syntaxErrors > 0:
    #     raise ast.ParseError("Unit {} had errors; terminating".format(name))
    return tree.transform(simplify.Simplify())


def translate(unit):
    with open_unit(unit) as input_file:
        obj = parse(input_file, name=unit.name)
        for p in passes:
            obj = obj.transform(p())
            logger.debug(__("Pass {}:\n{:p}", p.__name__, obj))

    archive = blum.Archive()
    for node in obj.select("toplevels", "stmt"):
        if node.t is 'fun_symbol':
            sym_name = '{}.{}'.format(unit.stem, node.attrs['name'])
            sym = blum.Symbol(
                section='text',
                data=node.attrs['text'],
                type_info=node.attrs['type'])
            archive.symbols[sym_name] = sym

    return archive



# jeff65 linker core functions
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import pathlib
import tempfile
from . import image


def link(name, archive, output_path):
    if not isinstance(output_path, pathlib.Path):
        output_path = pathlib.PurePath(output_path)

    # Create a temporary file to write to so that if we have a link error, we
    # don't clobber the existing program image. We create this in the same
    # directory to avoid cross-device renaming problems.
    im_fd, im_tmp = tempfile.mkstemp(prefix=output_path.name,
                                     dir=output_path.parent)
    try:
        with open(im_fd, 'wb') as im_file:
            im = image.Image(im_file)
            im.add_archive(image.make_startup_for(name, 0x0100))
            im.add_archive(archive)
            im.link()
        os.replace(im_tmp, output_path)
    except BaseException:
        # linking failed, remove the temporary file
        os.remove(im_tmp)
        raise



# jeff65 module root
# Copyright (C) 2017  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import logging
import pathlib
import sys


class BraceMessage:
    def __init__(self, fmt, *args, **kwargs):
        self.fmt = fmt
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        return self.fmt.format(*self.args, **self.kwargs)

    @classmethod
    def install(cls):
        """Add to builtins as __.

        Hacking builtins is generally a bad idea, but this is similar to _ in
        gettext, which is part of the standard library, so...
        """
        import builtins
        builtins.__ = cls


BraceMessage.install()


def main(argv=None):
    parser = argparse.ArgumentParser()

    parser.add_argument("--debug",
                        help="run in debug mode",
                        dest="debug", action="store_true",
                        default=False)
    parser.add_argument("-v", "--verbose",
                        help="show additional information during operation",
                        dest="verbose", action="store_true",
                        default=False)

    subparsers = parser.add_subparsers()
    compile_parser = subparsers.add_parser(
        'compile', help="compile one or more files")
    compile_parser.add_argument("-o", help="place the output into OUTPUT",
                                dest="output", type=pathlib.PurePath)
    compile_parser.add_argument('file', help="the file to compile",
                                type=pathlib.PurePath)
    compile_parser.set_defaults(func=cmd_compile)

    objdump_parser = subparsers.add_parser(
        'objdump', help="list symbols of an object")
    objdump_parser.add_argument('file', help="the file to examine",
                                type=pathlib.PurePath)
    objdump_parser.set_defaults(func=cmd_objdump)

    args = parser.parse_args(argv)
    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    elif args.verbose:
        logging.basicConfig(level=logging.INFO)

    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_help()
        sys.exit(1)


def cmd_none(args):
    raise NotImplementedError


def cmd_compile(args):
    from . import gold
    from . import blum

    archive = gold.translate(args.file)
    # archive.dumpf(args.file.with_suffix('.blum'))
    blum.link('{}.main'.format(args.file.stem), archive,
              args.output or args.file.with_suffix(".prg"))


def cmd_objdump(args):
    from . import blum

    archive = blum.Archive()
    archive.loadf(args.file)

    print("Unit: '{}'".format(args.file.stem))

    if len(archive.symbols) == 0:
        print("Symbols: (none)")
    else:
        print("Symbols:")
        for name, symbol in archive.symbols.items():
            print("    {}: size={}, section={}, type={}".format(
                name, len(symbol.data), symbol.section, symbol.type_info))
            for relocation in symbol.relocations:
                print("      {}".format(relocation))



# jeff65 AST manipulation
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import attr
from .immutable import FrozenDict


@attr.s(slots=True, frozen=True, repr=False)
class AstNode:
    t = attr.ib()
    attrs = attr.ib(factory=FrozenDict.empty, converter=FrozenDict.create)
    span = attr.ib(default=None, cmp=False)

    def update_attrs(self, attrs):
        nn = attr.evolve(self, attrs=self.attrs.update(attrs))
        assert self is not nn
        return nn

    def replace_attrs(self, attrs):
        nn = attr.evolve(self, attrs=FrozenDict.create(attrs))
        assert self is not nn
        return nn

    def transform(self, transformer):
        node = transformer.transform_enter(self.t, self)

        if hasattr(node, "attrs"):
            # because we're iterating over a FrozenDict, we'll get the
            # attributes in sorted order. Therefore, we can use the faster
            # unsafeset to put them into the builder.
            attrs = FrozenDict.Builder.empty()
            dirty = False
            for n, v in node.attrs.items():
                if hasattr(v, "transform"):
                    tv = v.transform(transformer)
                    attrs.unsafeset(n, tv)
                    if v is not tv:
                        dirty = True
                else:
                    attrs.unsafeset(n, v)
            if dirty:
                node = node.replace_attrs(attrs)

        return transformer.transform_exit(self.t, node)

    def __repr__(self):
        return f"<ast {self.t} at {self.span}>"

    def __format__(self, spec):
        if spec == 'p':
            return self.pretty()
        return repr(self)

    def pretty(self, indent=0, no_position=False):
        return self._pretty(indent, no_position).strip()

    def _pretty(self, indent, no_position):
        def i(n=0):
            return " " * (indent + n)

        pp = []

        if no_position:
            pp.append("{}{}\n".format(i(), self.t))
        else:
            pp.append("{}{:<{}} {}\n".format(i(), self.t, 70 - indent,
                                             self.span))
        has_next = False
        for name, value in self.attrs.items():
            if name == 'next':
                has_next = True
                continue  # we'll come back to this one
            if hasattr(value, "_pretty"):
                pp.append("{}:{} ".format(i(), name))
                pp.append(value._pretty(indent + 2 + len(name),
                                        no_position).lstrip())
            else:
                pp.append("{}:{} {!r}\n".format(i(), name, value))
        if has_next:
            # print this at the same level
            nxt = self.attrs['next']
            if hasattr(nxt, "_pretty"):
                pp.append(nxt._pretty(indent, no_position))
            elif nxt is not None:
                pp.append('{}{!r}'.format(i(), nxt))
        return "".join(pp)

    @classmethod
    def make_sequence(cls, t_seq, a_elem, elems, rest=None):
        n = rest
        for e in reversed(elems):
            n = cls(t_seq, {
                a_elem: e,
                'next': n,
            })
        return n

    def select(self, *attrs):
        current = [self]
        for a in attrs:
            found = []
            while len(current) > 0:
                c = current.pop()
                if c is None:
                    continue
                found.append(c.attrs[a])
                if "next" in c.attrs:
                    current.append(c.attrs["next"])
            current = found

        return current


class TranslationPass:
    """Base class for translation passes."""

    def transform_enter(self, t, node):
        return getattr(self, f'enter_{t}', self.__generic_enter)(node)

    def transform_exit(self, t, node):
        return getattr(self, f'exit_{t}', self.__generic_exit)(node)

    def __generic_enter(self, node):
        return node

    def __generic_exit(self, node):
        return node



# jeff65 disk format helper functions
# Copyright (C) 2017  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import struct


class Fmt:
    _bool = 0
    _u8 = 1
    _u16 = 2
    _u32 = 3
    _i8 = 4
    _i16 = 5
    _i32 = 6
    _str = 7
    _blob = 8
    _array = 10
    _table = 11
    _struct = 12
    _union = 13

    bool = (_bool,)  # pay attention to scope!
    u8 = (_u8,)
    u16 = (_u16,)
    u32 = (_u32,)
    i8 = (_i8,)
    i16 = (_i16,)
    i32 = (_i32,)
    str = (_str,)  # pay attention to scope!
    blob = (_blob,)

    @staticmethod
    def make_cc(code):
        cc, = struct.unpack('<H', code.encode('ascii'))
        return cc

    @classmethod
    def array(cls, item):
        return (cls._array, item)

    @classmethod
    def table(cls, key, value):
        return (cls._table, key, value)

    @classmethod
    def struct(cls, ty):
        return (cls._struct, ty)

    @classmethod
    def union(cls, tys):
        return (cls._union, tys)

    @classmethod
    def find_blobs(cls, obj):
        """Find all blobs for an object in traversal (DFS) order."""
        for name, fmt, _, pack in obj.fields:
            if not pack:
                continue

            # internal function so that we can recursively investigate arrays
            # and tables.
            def search(fmt, val):
                if fmt[0] == cls._blob:
                    yield val
                elif fmt[0] == cls._array:
                    for v in val:
                        yield from search(fmt[1], v)
                elif fmt[0] == cls._table:
                    for _, v in val.items():
                        yield from search(fmt[2], v)
                elif fmt[0] == cls._struct or fmt[0] == cls._union:
                    yield from cls.find_blobs(val)

            yield from search(fmt, getattr(obj, name))



# jeff65 type system
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import struct
from .fmt import Fmt


# We'll make this empty for now, and use it in class definitions, then mutate
# it at the end of the file, causing classes to "suddenly" know about the other
# classes.
known = []
fmt_type_info = Fmt.union(known)


class PhantomType:
    """An unreferenceable type."""

    discriminator = Fmt.make_cc('Ph')
    fields = []

    def __repr__(self):
        return '???'

    def __eq__(self, other):
        return isinstance(other, PhantomType)

    def validate(self):
        pass

    @classmethod
    def _empty(cls):
        return cls()


class VoidType:
    """A type with no values."""

    discriminator = Fmt.make_cc('Vd')
    fields = []

    def __repr__(self):
        return 'void'

    def __eq__(self, other):
        return isinstance(other, VoidType)

    def validate(self):
        pass

    @classmethod
    def _empty(cls):
        return cls()


class IntType:
    """An integral type."""

    discriminator = Fmt.make_cc('In')
    fields = [
        ('width', Fmt.u8, Fmt.make_cc('wd'), True),
        ('signed', Fmt.bool, Fmt.make_cc('sg'), True),
    ]

    def __init__(self, width, signed):
        self.width = width
        self.signed = signed

    def can_assign_from(self, rtype):
        return (type(rtype) is IntType
                and self.signed == rtype.signed
                and self.width >= rtype.width)

    def encode(self, value):
        if self.signed:
            return struct.pack('<q', value)
        return struct.pack('<Q', value)

    def decode(self, data):
        if self.signed:
            val, = struct.unpack('<q', data)
        else:
            val, = struct.unpack('<Q', data)
        assert val in self, "{} not in range for {}".format(val, repr(self))
        return val

    def __contains__(self, value):
        if not isinstance(value, int):
            return False

        bits = self.width * 8
        if self.signed:
            upper = (1 << (bits - 1))
            lower = -upper
        else:
            upper = (1 << bits)
            lower = 0
        return lower <= value < upper

    def __eq__(self, other):
        return (type(other) is IntType
                and self.signed == other.signed
                and self.width == other.width)

    def __repr__(self):
        return "{}{}".format(
            'i' if self.signed else 'u',
            self.width * 8)

    def validate(self):
        assert isinstance(self.width, int)
        assert self.width in i8
        assert isinstance(self.signed, bool)

    @classmethod
    def _empty(cls):
        return cls(None, None)


class RefType:
    """A reference type."""

    discriminator = Fmt.make_cc('Rf')
    fields = [
        ('target', fmt_type_info, Fmt.make_cc('tg'), True),
    ]

    def __init__(self, target):
        self.target = target
        self.width = 2

    def encode(self, value):
        return struct.pack('<H6x', value)

    def decode(self, data):
        val, = struct.unpack('<H6x', data)
        return val

    def can_assign_from(self, rtype):
        # Pointer assignment has to be equal unless the rhs is a mystery
        # pointer, in which case we just allow it to go through.
        return (self == rtype or
                (type(rtype) is RefType
                 and rtype.target is None))

    def __eq__(self, other):
        return (type(other) is RefType
                and self.target == other.target)

    def __repr__(self):
        return "&{}".format(repr(self.target))

    def validate(self):
        assert self.target is not None
        assert self.width == 2

    @classmethod
    def _empty(cls):
        return cls(None)


class FunctionType:
    """A function type."""

    discriminator = Fmt.make_cc('Fn')
    fields = [
        ('ret', fmt_type_info, Fmt.make_cc('rt'), True),
        ('args', Fmt.array(fmt_type_info), Fmt.make_cc('as'), True),
    ]

    def __init__(self, ret, *args):
        self.ret = ret
        self.args = list(args)
        self.width = None

    def can_assign_from(self, rtype):
        return self == rtype

    def __eq__(self, other):
        return (
            isinstance(other, FunctionType)
            and self.ret == other.ret
            and self.args == other.args)

    def __repr__(self):
        args = ", ".join(repr(arg) for arg in self.args)
        ret = ""
        if not isinstance(self.ret, VoidType):
            ret = " -> {}".format(repr(self.ret))
        return "fun({}){}".format(args, ret)

    def validate(self):
        assert self.ret is not None
        assert self.args is not None
        assert self.width is None

    @classmethod
    def _empty(cls):
        obj = cls(None)
        obj.args = None
        return obj


u8 = IntType(1, signed=False)
u16 = IntType(2, signed=False)
u24 = IntType(3, signed=False)
u32 = IntType(4, signed=False)
i8 = IntType(1, signed=True)
i16 = IntType(2, signed=True)
i24 = IntType(3, signed=True)
i32 = IntType(4, signed=True)
void = VoidType()
phantom = PhantomType()
ptr = RefType(phantom)

# Because we mutate this list, everywhere that uses it now knows about these
# types.
known.extend([
    PhantomType,
    VoidType,
    IntType,
    RefType,
    FunctionType,
])



# jeff64 gold-syntax CST -> AST simplification
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from ..grammar import T
from ... import ast, parsing, pattern
from ...pattern import Predicate as P


def require_token(t):
    return P.require(lambda n, c: n.t == t)


def token(t, key=None):
    return P(key, lambda n, c: n.t == t)


def unop(operator, sym, a_rhs):
    @pattern.match(
        ast.AstNode('expr', {
            "exhaustive!": True,
            "00": token(operator),
            "01": P('rhs'),
        }))
    def name_unop(self, rhs):
        return ast.AstNode(sym, {a_rhs: rhs})

    return name_unop


def binop(operator, sym):
    @pattern.match(
        ast.AstNode('expr', {
            "exhaustive!": True,
            "00": P("lhs"),
            "01": token(operator),
            "02": P("rhs"),
        }))
    def name_binop(self, lhs, rhs):
        return ast.AstNode(sym, {
            "lhs": lhs,
            "rhs": rhs,
        })

    return name_binop


def to_list(sym, t_node, a_element):
    @pattern.match(
        ast.AstNode(sym, {
            "exhaustive!": True,
            "00": P("car"),
            "01": P("cdr"),
        }))
    def name_right_recursion(self, car, cdr):
        return ast.AstNode(sym, {
            a_element: car,
            "next": cdr,
        })

    @pattern.match(ast.AstNode(sym, {"exhaustive!": True}))
    def name_right_recursion_final(self):
        return None

    return name_right_recursion, name_right_recursion_final


def drop_if_one_child(sym):
    @pattern.match(
        ast.AstNode(sym, {
            "exhaustive!": True,
            "00": P.any_node("inner"),
        }))
    def remove_outer(self, inner):
        return inner

    return remove_outer


@pattern.transform(pattern.Order.Ascending)
class Simplify:

    remove_outer_expr = drop_if_one_child('expr')
    remove_outer_alist = drop_if_one_child('alist')

    list_toplevel, list_toplevel_f = to_list("toplevel", "toplevel", "stmt")
    list_block, list_block_f = to_list("block", "block", "stmt")

    @pattern.match(
        ast.AstNode("unit", {
            "exhaustive!": True,
            "00": P("toplevels"),
        }))
    def name_unit(self, toplevels):
        return ast.AstNode("unit", {
            "toplevels": toplevels,
        })

    @pattern.match(ast.AstNode("alist", {"exhaustive!": True}))
    def remove_empty_alist(self):
        return None

    @pattern.match(
        ast.AstNode("alist_inner", {
            "exhaustive!": True,
            "00": P("arg"),
        }))
    def collapse_alist_final(self, arg):
        return ast.AstNode("alist", {
            "arg": arg,
            "next": None,
        })

    @pattern.match(
        ast.AstNode("alist_inner", {
            "exhaustive!": True,
            "00": P("arg"),
            "01": require_token(T.PUNCT_COMMA),
            "02": P("cdr"),
        }))
    def collapse_alist(self, arg, cdr):
        return ast.AstNode("alist", {
            "arg": arg,
            "next": cdr,
        })

    @pattern.match(
        ast.AstNode('stmt_use', {
            "exhaustive!": True,
            "00": require_token(T.STMT_USE),
            "01": P('unit_name'),
        }))
    def collapse_stmt_use(self, unit_name):
        return ast.AstNode('use', attrs={
            'name': unit_name.text,
        })

    @pattern.match(
        ast.AstNode('stmt_constant', {
            "exhaustive!": True,
            "00": require_token(T.STMT_CONSTANT),
            "01": ast.AstNode('declaration', {
                "exhaustive!": True,
                "00": P('name'),
                "01": require_token(T.PUNCT_COLON),
                "02": P('ty'),
            }),
            "02": require_token(T.OPERATOR_ASSIGN),
            "03": P('rhs'),
        }))
    def collapse_stmt_constant(self, name, ty, rhs):
        return ast.AstNode('constant', attrs={
            'name': name.text,
            'type': ty,
            'value': rhs,
        })

    @pattern.match(
        ast.AstNode("storage", {
            "exhaustive!": True,
        }))
    def storage_default(self):
        return None

    @pattern.match(
        ast.AstNode("storage", {
            "exhaustive!": True,
            "00": token(T.STORAGE_MUT),
        }))
    def storage_mut(self):
        return "mut"

    @pattern.match(
        ast.AstNode("storage", {
            "exhaustive!": True,
            "00": token(T.STORAGE_STASH),
        }))
    def storage_stash(self):
        return "stash"

    @pattern.match(
        ast.AstNode("stmt_let", {
            "exhaustive!": True,
            "00": require_token(T.STMT_LET),
            "01": P("storage"),
            "02": ast.AstNode("declaration", {
                "00": P("name"),
                "01": require_token(T.PUNCT_COLON),
                "02": P("ty"),
            }),
            "03": require_token(T.OPERATOR_ASSIGN),
            "04": P("value"),
        }))
    def collapse_stmt_let(self, storage, name, ty, value):
        return ast.AstNode("let", {
            "name": name.text,
            "type": ty,
            "storage": storage,
            "value": value,
        })

    @pattern.match(
        ast.AstNode('stmt_assign', {
            "exhaustive!": True,
            "00": P('lhs'),
            "01": require_token(T.OPERATOR_ASSIGN),
            "02": P('rhs'),
        }))
    def collapse_stmt_assign(self, lhs, rhs):
        return ast.AstNode('set', {
            "lvalue": lhs,
            "rvalue": rhs,
        })

    # # TODO: handle arguments, return values
    @pattern.match(
        ast.AstNode('stmt_fun', {
            "exhaustive!": True,
            "00": require_token(T.STMT_FUN),
            "01": P('name'),
            "02": require_token(T.PAREN_OPEN),
            "03": P.any(),  # plist
            "04": require_token(T.PAREN_CLOSE),
            "05": P("body"),
            "06": require_token(T.PUNCT_ENDFUN),
        }))
    def collapse_stmt_fun(self, name, body):
        return ast.AstNode('fun', {
            'name': name.text,
            'return': None,
            'args': None,
            'body': body,
        })

    @pattern.match(
        ast.AstNode('type_id', {
            "exhaustive!": True,
            "00": P('ty'),
        }))
    def simple_type(self, ty):
        return ty.text

    @pattern.match(
        ast.AstNode('type_id', {
            "exhaustive!": True,
            "00": token(T.OPERATOR_REF),
            "01": P("storage"),  # TODO figure out how to handle this properly
            "02": P('ty'),
        }))
    def ref_type(self, storage, ty):
        return ast.AstNode('type_ref', {
            'type': ty,
            # **{'storage': s.text for s in storage},
            'storage': storage,
        })

    @pattern.match(
        ast.AstNode('expr', {
            "exhaustive!": True,
            "00": token(T.NUMERIC, 'n'),
        }))
    def numeric(self, n):
        try:
            if n.text.startswith('0x'):
                value = int(n.text[2:], 16)
            elif n.text.startswith('0o'):
                value = int(n.text[2:], 8)
            elif n.text.startswith('0b'):
                value = int(n.text[2:], 2)
            else:
                value = int(n.text)
        except ValueError as e:
            raise parsing.ParseError(str(e))

        return ast.AstNode('numeric', {'value': value})

    @pattern.match(
        ast.AstNode('expr', {
            "exhaustive!": True,
            "00": token(T.IDENTIFIER, 'id'),
        }))
    def identifier(self, id):
        return ast.AstNode('identifier', {
            'name': id.text,
        })

    @pattern.match(
        ast.AstNode('expr', {
            "exhaustive!": True,
            "00": token(T.PAREN_OPEN),
            "01": P('inner'),
            "02": token(T.PAREN_CLOSE),
        }))
    def drop_expr_parens(self, inner):
        return inner

    name_negate = unop(T.OPERATOR_MINUS, 'negate', 'value')
    name_deref = unop(T.OPERATOR_DEREF, 'deref', 'address')
    name_add = binop(T.OPERATOR_PLUS, 'add')
    name_sub = binop(T.OPERATOR_MINUS, 'sub')
    name_mul = binop(T.OPERATOR_TIMES, 'mul')
    name_div = binop(T.OPERATOR_DIVIDE, 'div')

    @pattern.match(
        ast.AstNode('expr', {
            "exhaustive!": True,
            "00": P('namespace'),
            "01": token(T.OPERATOR_DOT),
            "02": ast.AstNode('member', {
                "exhaustive!": True,
                "00": P('member'),
            }),
        }))
    def name_member_access(self, namespace, member):
        return ast.AstNode('member_access', attrs={
            'namespace': namespace,
            'member': member.text,
        })

    @pattern.match(
        ast.AstNode('stmt_call', {
            "exhaustive!": True,
            "00": P('function'),
            "01": token(T.PAREN_OPEN),
            "02": P('alist'),  # TODO handle this properly
            "03": token(T.PAREN_CLOSE),
        }))
    def name_call(self, function, alist):
        return ast.AstNode('call', attrs={
            'target': function,
            'args': alist,
        })

    # Collapse left-recursion on strings.
    @pattern.match(ast.AstNode('string_inner', {"exhaustive!": True}))
    def string_inner_empty(self):
        return []

    @pattern.match(
        ast.AstNode('string_inner', {
            "00": P('string0'),
            "01": token(T.STRING, 'string1'),
        }))
    def string_inner_segment(self, string0, string1):
        string0.append(string1.text)
        return string0

    @pattern.match(
        ast.AstNode('string_inner', {
            "00": P('string0'),
            "01": token(T.STRING_ESCAPE, 'string1'),
        }))
    def string_inner_escape(self, string0, string1):
        string0.append(string1.text[1])
        return string0

    @pattern.match(
        ast.AstNode('string', {
            "00": require_token(T.STRING_DELIM),
            "01": P('value'),
            "02": require_token(T.STRING_DELIM),
        }))
    def collapse_string(self, value):
        return ast.AstNode('string', attrs={'value': "".join(value)})



# jeff65 gold-syntax name binding
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from ... import ast
from ...immutable import FrozenDict


class ScopedPass(ast.TranslationPass):
    """Base class for translation passes which understand binding scope.
    """
    scoped_types = ['unit', 'fun']

    def __init__(self):
        self.scopes = []

    def bind_name(self, name, value):
        self.scopes[-1]['known_names'][name] = value

    def look_up_name(self, name):
        for scope in reversed(self.scopes):
            known_names = scope['known_names']
            if name in known_names:
                return known_names[name]
        return None

    def bind_constant(self, name, value):
        self.scopes[-1]['known_constants'][name] = value

    def look_up_constant(self, name):
        for scope in reversed(self.scopes):
            known_constants = scope['known_constants']
            if name in known_constants:
                return known_constants[name]
        return None

    def transform_enter(self, t, node):
        node = super().transform_enter(t, node)
        if t in self.scoped_types:
            self.scopes.append({
                "known_names": node.attrs.get(
                    "known_names", FrozenDict.empty()).asbuilder(),
                "known_constants": node.attrs.get(
                    "known_constants", FrozenDict.empty()).asbuilder(),
            })
            node = self.enter__scope(node)
        return node

    def transform_exit(self, t, node):
        if t in self.scoped_types:
            node = self.exit__scope(node)
        if t in self.scoped_types:
            node = node.update_attrs(
                {k: v.asfrozen() for k, v in self.scopes.pop().items()})
        node = super().transform_exit(t, node)
        return node

    def enter__scope(self, node):
        return node

    def exit__scope(self, node):
        return node


class ShadowNames(ScopedPass):
    """Binds names to dummy values, to be overridden later.

    This allows us to determine whether module names are shadowed while
    constructing types.
    """

    def exit_constant(self, node):
        self.bind_name(node.attrs['name'], True)
        return node


class BindNamesToTypes(ScopedPass):
    """Binds names to types. These are later overridden by the storage."""

    def exit_constant(self, node):
        self.bind_name(node.attrs['name'], node.attrs['type'])
        return node


class EvaluateConstants(ScopedPass):
    def __init__(self):
        super().__init__()
        self.evaluating = False

    def enter_constant(self, node):
        self.evaluating = True
        return node

    def exit_constant(self, node):
        self.evaluating = False
        self.bind_constant(node.attrs['name'], node.attrs["value"])
        return None

    def exit_toplevel(self, node):
        if node.attrs["stmt"] is None:
            return node.attrs["next"]
        return node

    def exit_call(self, node):
        target = node.attrs['target']
        return target(*node.select("args", "arg"))


class ResolveConstants(ScopedPass):
    def exit_identifier(self, node):
        value = self.look_up_constant(node.attrs['name'])
        if not value:
            return node
        return value



# jeff65 immutable datatypes
# Copyright (C) 2018  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import bisect
import collections.abc


class FrozenDict(collections.abc.Mapping):
    """An immutable mapping type.

    Currently this is implemented as a sorted list, but in the future it could
    be replaced with a more efficient implementation if it proves to be slow.
    """

    def __init__(self, keys, values):
        self.__keys = tuple(keys)
        self.__values = tuple(values)

    def __repr__(self):
        pairs = ", ".join(f'{k!r}: {v!r}' for k, v in self.items())
        return f'FrozenDict({{{pairs}}})'

    def __iter__(self):
        return iter(self.__keys)

    def __len__(self):
        return len(self.__keys)

    def __getitem__(self, key):
        k = bisect.bisect_left(self.__keys, key)
        if k != len(self.__keys) and self.__keys[k] == key:
            return self.__values[k]
        raise KeyError(key)

    def add(self, key, value):
        builder = self.asbuilder()
        builder[key] = value
        return builder.asfrozen()

    def update(self, mapping):
        builder = self.asbuilder()
        builder.update(mapping)
        return builder.asfrozen()

    def remove(self, key):
        builder = self.asbuilder()
        del builder[key]
        return builder.asfrozen()

    def asbuilder(self):
        return self.Builder(self.__keys, self.__values)

    @classmethod
    def empty(cls):
        return cls((), ())

    @classmethod
    def create(cls, other):
        if isinstance(other, cls):
            return other  # No need to re-create!
        elif isinstance(other, cls.Builder):
            return other.asfrozen()

        builder = cls.Builder.empty()
        builder.update(other)
        return builder.asfrozen()

    class Builder(collections.abc.MutableMapping):
        def __init__(self, keys, values):
            self.__keys = list(keys)
            self.__values = list(values)

        def __repr__(self):
            pairs = ", ".join(f'{k!r}: {v!r}' for k, v in self.items())
            return f'FrozenDict.Builder({{{pairs}}})'

        def __iter__(self):
            return iter(self.__keys)

        def __len__(self):
            return len(self.__keys)

        def __getitem__(self, key):
            k = bisect.bisect_left(self.__keys, key)
            if k != len(self.__keys) and self.__keys[k] == key:
                return self.__values[k]
            raise KeyError(key)

        def __setitem__(self, key, value):
            k = bisect.bisect_left(self.__keys, key)
            if k != len(self.__keys) and self.__keys[k] == key:
                self.__keys[k] = key
                self.__values[k] = value
            else:
                self.__keys.insert(k, key)
                self.__values.insert(k, value)

        def __delitem__(self, key):
            k = bisect.bisect_left(self.__keys, key)
            if k != len(self.__keys) and self.__keys[k] == key:
                del self.__keys[k]
                del self.__values[k]
            else:
                raise KeyError(key)

        def unsafeset(self, key, value):
            self.__keys.append(key)
            self.__values.append(value)

        def asfrozen(self):
            return FrozenDict(self.__keys, self.__values)

        @classmethod
        def empty(cls):
            return cls([], [])



# jeff65 gold-syntax resolution passes
# Copyright (C) 2017  jeff65 maintainers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from . import binding
from .. import mem
from ... import ast, pattern
from ...pattern import Predicate as P


@pattern.transform(pattern.Order.Descending)
class ResolveStorage:
    @pattern.match(
        ast.AstNode('deref', attrs={
            'type': P('ty'),
            'address': ast.AstNode(P.require('numeric'), attrs={
                'value': P('address'),
            }),
        }))
    def deref_to_absolute(self, ty, address):
        return ast.AstNode('absolute_storage', attrs={
            'address': address,
            'width': ty.width
        })

    @pattern.match(
        ast.AstNode('numeric', attrs={
            'value': P.lt(256, 'value', require=True),
        }))
    def numeric_to_immediate(self, value):
        return ast.AstNode('immediate_storage', attrs={
            'value': value,
            'width': 1,
        })


class ResolveUnits(binding.ScopedPass):
    """Resolves external units identified in 'use' statements."""

    builtin_units = {
        'mem': mem.MemUnit(),
    }

    def exit_use(self, node):
        name = node.attrs['name']
        unit = self.builtin_units[name]
        self.bind_name(name, unit)
        return None

    def exit_toplevel(self, node):
        if node.attrs["stmt"] is None:
            return node.attrs["next"]
        return node


class ResolveMembers(binding.ScopedPass):
    """Resolves members to functions."""

    def exit_member_access(self, node):
        member = node.attrs['member']
        name = node.attrs['namespace'].attrs['name']
        unit = self.look_up_name(name)
        return unit.member(member)
